# coding=utf-8

import asyncio

import pytest
import async_timeout

from xpaw.queue import FifoQueue, LifoQueue, PriorityQueue
from xpaw.config import Config
from xpaw.eventbus import EventBus
from xpaw.http import HttpRequest


class Cluster:
    def __init__(self, loop=None, **kwargs):
        self.loop = loop
        self.config = Config(kwargs)
        self.event_bus = EventBus()


async def test_fifo_queue(loop):
    q = FifoQueue.from_cluster(Cluster(loop=loop))
    with pytest.raises(asyncio.TimeoutError):
        with async_timeout.timeout(0.1, loop=loop):
            await q.pop()
    obj_list = [HttpRequest('1'), HttpRequest('2'), HttpRequest('3')]
    for o in obj_list:
        await q.push(o)
    for i in range(len(obj_list)):
        assert await q.pop() == obj_list[i]
    with pytest.raises(asyncio.TimeoutError):
        with async_timeout.timeout(0.1, loop=loop):
            await q.pop()


async def test_fifo_queue_dump(loop, tmpdir):
    q = FifoQueue.from_cluster(Cluster(loop=loop, dump_dir=str(tmpdir)))
    obj_list = [HttpRequest('1'), HttpRequest('2'), HttpRequest('3')]
    for o in obj_list:
        await q.push(o)
    await q.close()
    q2 = FifoQueue.from_cluster(Cluster(loop=loop, dump_dir=str(tmpdir)))
    await q2.open()
    with async_timeout.timeout(0.1, loop=loop):
        for i in range(len(obj_list)):
            assert (await q2.pop()).url == obj_list[i].url


async def test_lifo_queue(loop):
    q = LifoQueue.from_cluster(Cluster(loop=loop))
    with pytest.raises(asyncio.TimeoutError):
        with async_timeout.timeout(0.1, loop=loop):
            await q.pop()
    obj_list = [HttpRequest('1'), HttpRequest('2'), HttpRequest('3')]
    for o in obj_list:
        await q.push(o)
    for i in range(len(obj_list)):
        assert await q.pop() == obj_list[len(obj_list) - i - 1]
    with pytest.raises(asyncio.TimeoutError):
        with async_timeout.timeout(0.1, loop=loop):
            await q.pop()


async def test_lifo_queue_dump(loop, tmpdir):
    q = LifoQueue.from_cluster(Cluster(loop=loop, dump_dir=str(tmpdir)))
    obj_list = [HttpRequest('1'), HttpRequest('2'), HttpRequest('3')]
    for o in obj_list:
        await q.push(o)
    await q.close()
    q2 = LifoQueue.from_cluster(Cluster(loop=loop, dump_dir=str(tmpdir)))
    await q2.open()
    with async_timeout.timeout(0.1, loop=loop):
        for i in range(len(obj_list)):
            assert (await q2.pop()).url == obj_list[len(obj_list) - i - 1].url


async def test_priority_queue(loop):
    item1_1 = HttpRequest('1_1', priority=1)
    item1_2 = HttpRequest('1_2', priority=1)
    item2_1 = HttpRequest('2_1', priority=2)
    item2_2 = HttpRequest('2_2', priority=2)
    item3_1 = HttpRequest('3_1', priority=3)
    item3_2 = HttpRequest('3_2', priority=3)
    q = PriorityQueue.from_cluster(Cluster(loop=loop))
    with pytest.raises(asyncio.TimeoutError):
        with async_timeout.timeout(0.1, loop=loop):
            await q.pop()
    await q.push(item2_1)
    await q.push(item1_1)
    await q.push(item3_1)
    await q.push(item1_2)
    await q.push(item2_2)
    await q.push(item3_2)
    assert await q.pop() is item3_1
    assert await q.pop() is item3_2
    assert await q.pop() is item2_1
    assert await q.pop() is item2_2
    assert await q.pop() is item1_1
    assert await q.pop() is item1_2
    with pytest.raises(asyncio.TimeoutError):
        with async_timeout.timeout(0.1, loop=loop):
            await q.pop()


async def test_priority_queue_dump(loop, tmpdir):
    q = PriorityQueue.from_cluster(Cluster(loop=loop, dump_dir=str(tmpdir)))
    await q.push(HttpRequest('2', priority=2))
    await q.push(HttpRequest('3', priority=3))
    await q.push(HttpRequest('1', priority=1))
    await q.close()
    q2 = PriorityQueue.from_cluster(Cluster(loop=loop, dump_dir=str(tmpdir)))
    await q2.open()
    with async_timeout.timeout(0.1, loop=loop):
        assert (await q2.pop()).url == '3'
        assert (await q2.pop()).url == '2'
        assert (await q2.pop()).url == '1'



# coding=utf-8

from xpaw import make_requests, HttpRequest

if __name__ == '__main__':
    requests = ['http://localhost', 'http://python.org', HttpRequest('http://python.org')]
    results = make_requests(requests)
    print(results)



# coding=utf-8

import logging

from .utils import load_object
from . import events
from .errors import NotEnabled

log = logging.getLogger(__name__)


class MiddlewareManager:
    def __init__(self, *middlewares):
        self._open_handlers = []
        self._close_handlers = []
        self.components = []
        for middleware in middlewares:
            self._add_middleware(middleware)

    @classmethod
    def _middleware_list_from_config(cls, config):
        raise NotImplementedError

    @classmethod
    def from_cluster(cls, cluster):
        mw_list = cls._middleware_list_from_config(cluster.config)
        mws = []
        for cls_path in mw_list:
            mw_cls = load_object(cls_path)
            try:
                if hasattr(mw_cls, "from_cluster"):
                    mw = mw_cls.from_cluster(cluster)
                else:
                    mw = mw_cls()
            except NotEnabled:
                log.debug('%s is not enabled', cls_path)
            else:
                mws.append(mw)
        mwm = cls(*mws)
        cluster.event_bus.subscribe(mwm.open, events.cluster_start)
        cluster.event_bus.subscribe(mwm.close, events.cluster_shutdown)
        return mwm

    def _add_middleware(self, middleware):
        self.components.append(middleware)
        if hasattr(middleware, "open"):
            self._open_handlers.append(middleware.open)
        if hasattr(middleware, "close"):
            self._close_handlers.insert(0, middleware.close)

    @staticmethod
    def _priority_list_from_config(name, config):
        c = config.get(name)
        assert c is None or isinstance(c, (list, dict)), \
            "'{}' must be None, a list or a dict, got {}".format(name, type(c).__name__)
        if c is None:
            return {}
        d = {}
        if isinstance(c, dict):
            for i in c:
                d[i] = (float(c[i]),)
        elif isinstance(c, list):
            for i in range(len(c)):
                d[c[i]] = (0, i)
        return d

    @classmethod
    def _make_component_list(cls, name, config):
        c_base = cls._priority_list_from_config(name + '_base', config)
        c = cls._priority_list_from_config(name, config)
        c_base.update(c)
        res = []
        for k, v in c_base.items():
            if v is not None:
                res.append((k, v))
        res.sort(key=lambda x: x[1])
        return [i[0] for i in res]

    def open(self):
        for method in self._open_handlers:
            method()

    def close(self):
        for method in self._close_handlers:
            method()



# coding=utf-8

import logging
import inspect

from .middleware import MiddlewareManager

log = logging.getLogger(__name__)


class ItemPipelineManager(MiddlewareManager):
    def __init__(self, *middlewares):
        self._item_handlers = []
        super().__init__(*middlewares)

    @classmethod
    def _middleware_list_from_config(cls, config):
        return cls._make_component_list('item_pipelines', config)

    def _add_middleware(self, middleware):
        super()._add_middleware(middleware)
        if hasattr(middleware, "handle_item"):
            self._item_handlers.append(middleware.handle_item)

    async def handle_item(self, item):
        log.debug('Item (%s): %s', type(item).__name__, item)
        for method in self._item_handlers:
            res = method(item)
            if inspect.iscoroutine(res):
                await res



# coding=utf-8

import logging
import os
from os.path import join, isfile
import sys
import asyncio
import signal

from .config import BaseConfig, Config
from .cluster import LocalCluster
from . import utils
from .spider import RequestsSpider

log = logging.getLogger(__name__)


def run_crawler(proj_dir, **kwargs):
    config = BaseConfig(kwargs)
    run_cluster(proj_dir=proj_dir, base_config=config)


def run_spider(spider, **kwargs):
    config = BaseConfig(kwargs)
    config.set("spider", spider)
    run_cluster(base_config=config)


def run_cluster(proj_dir=None, base_config=None):
    config = load_job_config(proj_dir=proj_dir, base_config=base_config)
    utils.configure_logger('xpaw', config)
    if config.getbool('daemon'):
        utils.be_daemon()
    pid_file = config.get('pid_file')
    _write_pid_file(pid_file)
    try:
        loop = _get_event_loop()
        cluster = LocalCluster(config, loop=loop)
    except Exception:
        log.error('Failed to create cluster', exc_info=True)
        _remove_pid_file(pid_file)
        utils.remove_logger('xpaw')
        raise
    default_signal_handlers = _set_signal_handlers(cluster)
    try:
        loop.run_until_complete(cluster.run())
    finally:
        _remove_pid_file(pid_file)
        _recover_signal_handlers(default_signal_handlers)
        utils.remove_logger('xpaw')


def make_requests(requests, **kwargs):
    if 'log_level' not in kwargs:
        kwargs['log_level'] = 'WARNING'
    start_requests = [r for r in requests]
    results = [None] * len(start_requests)
    run_spider(RequestsSpider, start_requests=start_requests, results=results, **kwargs)
    return results


def _get_event_loop():
    try:
        loop = asyncio.get_event_loop()
    except RuntimeError:
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
    return loop


def load_job_config(proj_dir=None, base_config=None):
    if proj_dir is not None and proj_dir not in sys.path:
        # add project path
        sys.path.append(proj_dir)
    job_config = Config()
    if proj_dir is not None:
        config_file = join(proj_dir, 'config.py')
        if isfile(config_file):
            try:
                c = utils.load_config(config_file)
            except Exception:
                raise RuntimeError('Cannot read the configuration file {}'.format(config_file))
            for k, v in utils.iter_settings(c):
                job_config.set(k, v)
    job_config.update(base_config)
    return job_config


def _write_pid_file(pid_file):
    if pid_file is not None:
        with open(pid_file, 'w') as f:
            f.write(str(os.getpid()))


def _remove_pid_file(pid_file):
    if pid_file is not None:
        try:
            os.remove(pid_file)
        except Exception as e:
            log.warning('Cannot remove PID file %s: %s', pid_file, e)


def _set_signal_handlers(cluster):
    def _exit(signum, frame):
        log.info('Received exit signal: %s', signum)
        cluster.loop.call_soon_threadsafe(cluster.stop)

    default_signal_handlers = [(signal.SIGINT, signal.getsignal(signal.SIGINT)),
                               (signal.SIGTERM, signal.getsignal(signal.SIGTERM))]
    signal.signal(signal.SIGINT, _exit)
    signal.signal(signal.SIGTERM, _exit)
    return default_signal_handlers


def _recover_signal_handlers(handlers):
    for h in handlers:
        signal.signal(h[0], h[1])



# coding=utf-8

import copy
from collections import MutableMapping
import inspect


class BaseConfig(MutableMapping):
    def __init__(self, values=None):
        self.attributes = {}
        self.update(values)

    def __getitem__(self, opt_name):
        if opt_name not in self:
            return None
        return self.attributes[opt_name]

    def __contains__(self, name):
        return name in self.attributes

    def get(self, name, default=None):
        return self[name] if self[name] is not None else default

    def getbool(self, name, default=None):
        v = self.get(name, default)
        return getbool(v)

    def getint(self, name, default=None):
        v = self.get(name, default)
        return getint(v)

    def getfloat(self, name, default=None):
        v = self.get(name, default)
        return getfloat(v)

    def getlist(self, name, default=None):
        v = self.get(name, default)
        return getlist(v)

    def __setitem__(self, name, value):
        self.set(name, value)

    def set(self, name, value):
        self.attributes[name] = value

    def update(self, values):
        if values is not None:
            if isinstance(values, BaseConfig):
                for name in values:
                    self.set(name, values[name])
            else:
                for name, value in values.items():
                    self.set(name, value)

    def delete(self, name):
        del self.attributes[name]

    def __delitem__(self, name):
        del self.attributes[name]

    def copy(self):
        return copy.deepcopy(self)

    def __iter__(self):
        return iter(self.attributes)

    def __len__(self):
        return len(self.attributes)


def getbool(v):
    try:
        return bool(int(v))
    except (ValueError, TypeError):
        if v in ("True", "true"):
            return True
        if v in ("False", "false"):
            return False
    return None


def getint(v):
    try:
        return int(v)
    except (ValueError, TypeError):
        pass
    return None


def getfloat(v):
    try:
        return float(v)
    except (ValueError, TypeError):
        pass
    return None


def getlist(v):
    if v is None:
        return None
    if isinstance(v, str):
        v = v.split(",")
    elif not hasattr(v, "__iter__"):
        v = [v]
    return list(v)


class Config(BaseConfig):
    def __init__(self, values=None):
        super().__init__()
        for v in KNOWN_SETTINGS.values():
            self.set(v.name, v.value)
        self.update(values)


class Setting:
    name = None
    cli = None
    metavar = None
    default = None
    action = None
    type = None
    nargs = None
    short_desc = None

    def __init__(self):
        self.value = self.default

    def add_argument(self, parser):
        if self.cli is None:
            return
        args = tuple(self.cli)
        kwargs = {'dest': self.name, 'help': '{} (default: {})'.format(self.short_desc, self.default)}
        if self.metavar is not None:
            kwargs['metavar'] = self.metavar
        if self.action is not None:
            kwargs['action'] = self.action
        if self.type is not None:
            kwargs['type'] = self.type
        if self.nargs is not None:
            kwargs['nargs'] = self.nargs

        parser.add_argument(*args, **kwargs)


class Daemon(Setting):
    name = 'daemon'
    cli = ['-d', '--daemon']
    action = 'store_true'
    default = False
    short_desc = 'run in daemon mode'


class PidFile(Setting):
    name = 'pid_file'
    cli = ['--pid-file']
    metavar = 'FILE'
    short_desc = 'PID file'


class DumpDir(Setting):
    name = 'dump_dir'
    cli = ['--dump-dir']
    metavar = 'DIR'
    short_desc = 'the directory to dump the state of a single job'


class LogFile(Setting):
    name = 'log_file'
    cli = ['--log-file']
    metavar = 'FILE'
    short_desc = 'log file'


class LogLevel(Setting):
    name = 'log_level'
    cli = ['-l', '--log-level']
    metavar = 'LEVEL'
    default = 'INFO'
    short_desc = 'log level: DEBUG, INFO, WARNING, ERROR'


class LogFormat(Setting):
    name = 'log_format'
    default = '%(asctime)s %(name)s [%(levelname)s] %(message)s'


class LogDateformat(Setting):
    name = 'log_dateformat'
    default = '%Y-%m-%d %H:%M:%S'


class DownloaderClients(Setting):
    name = 'downloader_clients'
    cli = ['--downloader-clients']
    metavar = 'INT'
    type = int
    default = 100
    short_desc = 'the maximum number of simultaneous clients'


class DownloaderTimeout(Setting):
    name = 'downloader_timeout'
    cli = ['--downloader-timeout']
    metavar = 'FLOAT'
    type = float
    default = 20
    short_desc = 'timeout of downloader in seconds'


class VerifySsl(Setting):
    name = 'verify_ssl'
    default = False


class AllowRedirects(Setting):
    name = 'allow_redirects'
    default = True


class CookieJarEnabled(Setting):
    name = 'cookie_jar_enabled'
    cli = ['--cookie-jar-enabled']
    action = 'store_true'
    default = False
    short_desc = 'enable cookie jar'


class DefaultHeaders(Setting):
    name = 'default_headers'
    default = {
        'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,image/apng,*/*;q=0.8',
        'Accept-Encoding': 'gzip, deflate',
        'Accept-Language': 'zh-CN,zh;q=0.8',
        'Cache-Control': 'max-age=0',
        'Connection': 'keep-alive',
        'Upgrade-Insecure-Requests': '1'
    }


class UserAgent(Setting):
    name = 'user_agent'
    default = ':desktop'


class RandomUserAgent(Setting):
    name = 'random_user_agent'
    default = False


class ImitatingProxyEnabled(Setting):
    name = 'imitating_proxy_enabled'
    default = False


class Proxy(Setting):
    name = 'proxy'


class RetryEnabled(Setting):
    name = 'retry_enabled'
    default = True


class MaxRetryTimes(Setting):
    name = 'max_retry_times'
    default = 3


class RetryHttpStatus(Setting):
    name = 'retry_http_status'
    default = (500, 502, 503, 504, 408, 429)


class SpeedLimitEnabled(Setting):
    name = 'speed_limit_enabled'
    default = False


class SpeedLimitRate(Setting):
    name = 'speed_limit_rate'
    default = 1


class SpeedLimitBurst(Setting):
    name = 'speed_limit_burst'
    default = 1


class MaxDepth(Setting):
    name = 'max_depth'
    cli = ['--max-depth']
    metavar = 'INT'
    type = int
    short_desc = 'maximum depth of spider'


class StatsCenter(Setting):
    name = 'stats_collector'
    default = 'xpaw.stats.StatsCollector'


class Queue(Setting):
    name = 'queue'
    default = 'xpaw.queue.PriorityQueue'


class DupeFilter(Setting):
    name = 'dupe_filter'
    default = 'xpaw.dupefilter.HashDupeFilter'


class Spider(Setting):
    name = 'spider'


class DownloaderMiddlewares(Setting):
    name = 'downloader_middlewares'


class DownloaderMiddlewaresBase(Setting):
    name = 'downloader_middlewares_base'
    default = {
        # cluster side
        'xpaw.downloadermws.DefaultHeadersMiddleware': 300,
        'xpaw.downloadermws.ImitatingProxyMiddleware': 350,
        'xpaw.downloadermws.UserAgentMiddleware': 400,
        'xpaw.downloadermws.RetryMiddleware': 500,
        'xpaw.downloadermws.CookiesMiddleware': 600,
        'xpaw.downloadermws.ProxyMiddleware': 700,
        'xpaw.downloadermws.SpeedLimitMiddleware': 900,
        # downloader side
    }


class SpiderMiddlewares(Setting):
    name = 'spider_middlewares'


class SpiderMiddlewaresBase(Setting):
    name = 'spider_middlewares_base'
    default = {
        # cluster side
        'xpaw.spidermws.DepthMiddleware': 900
        # spider side
    }


class ItemPipelines(Setting):
    name = 'item_pipelines'


class Extensions(Setting):
    name = 'extensions'


KNOWN_SETTINGS = {}

for _v in list(vars().values()):
    if inspect.isclass(_v) and issubclass(_v, Setting) and _v.name is not None:
        KNOWN_SETTINGS[_v.name] = _v()



# coding=utf-8

import asyncio

TimeoutError = asyncio.TimeoutError


class NotEnabled(Exception):
    """
    Not enabled.
    """


class ClientError(Exception):
    """
    Downloader client error.
    """


class IgnoreRequest(Exception):
    """
    Ignore this request.
    """


class IgnoreItem(Exception):
    """
    Ignore this item.
    """


class UsageError(Exception):
    """
    CLI usage error.
    """


class StopCluster(Exception):
    """
    Stop cluster.
    """



# coding=utf-8

import logging
import asyncio
import inspect
from asyncio import CancelledError
from urllib.parse import urlsplit

import aiohttp
import async_timeout
from yarl import URL
from multidict import MultiDict
from aiohttp.helpers import BasicAuth

from .middleware import MiddlewareManager
from .http import HttpRequest, HttpResponse
from .errors import ClientError, TimeoutError

log = logging.getLogger(__name__)


class Downloader:
    def __init__(self, timeout=None, verify_ssl=False, allow_redirects=True, loop=None):
        self._timeout = timeout
        self._verify_ssl = verify_ssl
        self._allow_redirects = allow_redirects
        self._loop = loop or asyncio.get_event_loop()

    async def download(self, request):
        log.debug("HTTP request: %s", request)
        timeout = request.meta.get("timeout")
        if timeout is None:
            timeout = self._timeout
        verify_ssl = request.meta.get('verify_ssl')
        if verify_ssl is None:
            verify_ssl = self._verify_ssl
        allow_redirects = request.meta.get('allow_redirects')
        if allow_redirects is None:
            allow_redirects = self._allow_redirects
        cookie_jar = request.meta.get('cookie_jar')
        auth = parse_request_auth(request.meta.get('auth'))
        proxy = parse_request_url(request.meta.get('proxy'))
        proxy_auth = parse_request_auth(request.meta.get('proxy_auth'))
        async with aiohttp.ClientSession(connector=aiohttp.TCPConnector(verify_ssl=verify_ssl, loop=self._loop),
                                         cookies=request.cookies,
                                         cookie_jar=cookie_jar,
                                         loop=self._loop) as session:
            with async_timeout.timeout(timeout, loop=self._loop):
                if isinstance(request.body, dict):
                    data, json = None, request.body
                else:
                    data, json = request.body, None
                async with session.request(request.method,
                                           parse_request_url(request.url),
                                           params=parse_request_params(request.params),
                                           auth=auth,
                                           headers=request.headers,
                                           data=data,
                                           json=json,
                                           proxy=proxy,
                                           proxy_auth=proxy_auth,
                                           allow_redirects=allow_redirects) as resp:
                    body = await resp.read()
                    cookies = resp.cookies
        response = HttpResponse(resp.url,
                                resp.status,
                                headers=resp.headers,
                                body=body,
                                cookies=cookies)
        log.debug("HTTP response: %s", response)
        return response


class DownloaderMiddlewareManager(MiddlewareManager):
    def __init__(self, *middlewares):
        self._request_handlers = []
        self._response_handlers = []
        self._error_handlers = []
        super().__init__(*middlewares)

    def _add_middleware(self, middleware):
        super()._add_middleware(middleware)
        if hasattr(middleware, "handle_request"):
            self._request_handlers.append(middleware.handle_request)
        if hasattr(middleware, "handle_response"):
            self._response_handlers.insert(0, middleware.handle_response)
        if hasattr(middleware, "handle_error"):
            self._error_handlers.insert(0, middleware.handle_error)

    @classmethod
    def _middleware_list_from_config(cls, config):
        return cls._make_component_list('downloader_middlewares', config)

    async def download(self, downloader, request):
        try:
            res = await self._handle_request(request)
            if isinstance(res, HttpRequest):
                return res
            if res is None:
                try:
                    response = await downloader.download(request)
                except CancelledError:
                    raise
                except asyncio.TimeoutError:
                    log.debug('%s is timeout', request)
                    raise TimeoutError('Request timeout')
                except Exception as e:
                    log.debug("%s is error: %s", request, e)
                    raise ClientError(e)
                else:
                    res = response
        except CancelledError:
            raise
        except Exception as e:
            res = await self._handle_error(request, e)
            if isinstance(res, Exception):
                raise res
        if isinstance(res, HttpResponse):
            _res = await self._handle_response(request, res)
            if _res:
                res = _res
            # bind request
            res.request = request
        return res

    async def _handle_request(self, request):
        for method in self._request_handlers:
            res = method(request)
            if inspect.iscoroutine(res):
                res = await res
            assert res is None or isinstance(res, (HttpRequest, HttpResponse)), \
                "Request handler must return None, HttpRequest or HttpResponse, got {}".format(type(res).__name__)
            if res:
                return res

    async def _handle_response(self, request, response):
        for method in self._response_handlers:
            res = method(request, response)
            if inspect.iscoroutine(res):
                res = await res
            assert res is None or isinstance(res, HttpRequest), \
                "Response handler must return None or HttpRequest, got {}".format(type(res).__name__)
            if res:
                return res

    async def _handle_error(self, request, error):
        for method in self._error_handlers:
            res = method(request, error)
            if inspect.iscoroutine(res):
                res = await res
            assert res is None or isinstance(res, (HttpRequest, HttpResponse)), \
                "Exception handler must return None, HttpRequest or HttpResponse, got {}".format(type(res).__name__)
            if res:
                return res
        return error


def parse_request_params(params):
    if isinstance(params, dict):
        res = MultiDict()
        for k, v in params.items():
            if isinstance(v, (tuple, list)):
                for i in v:
                    res.add(k, i)
            else:
                res.add(k, v)
        params = res
    return params


def parse_request_auth(auth):
    if isinstance(auth, (tuple, list)):
        auth = BasicAuth(*auth)
    elif isinstance(auth, str):
        auth = BasicAuth(*auth.split(':', 1))
    return auth


def parse_request_url(url):
    if isinstance(url, str):
        res = urlsplit(url)
        if res.scheme == '':
            url = 'http://{}'.format(url)
        url = URL(url)
    return url



# coding=utf-8

text = '''
<html>
<head>
    <title>Quotes to Scrape</title>
    <link rel="stylesheet" href="/static/main.css">
</head>
<body>
    <div class="quote" itemscope itemtype="http://schema.org/CreativeWork">
        <span class="text" itemprop="text">“I have not failed. I&#39;ve just found 10,000 ways that won&#39;t work.”</span>
        <span>by <small class="author" itemprop="author">Thomas A. Edison</small>
        <a href="/author/Thomas-A-Edison">(about)</a>
        </span>
        <div class="tags">
            Tags:
            <a class="tag" href="/tag/edison/page/1/">edison</a>
            <a class="tag" href="/tag/failure/page/1/">failure</a>
            <a class="tag" href="/tag/inspirational/page/1/">inspirational</a>
            <a class="tag" href="/tag/paraphrased/page/1/">paraphrased</a>
        </div>
    </div>
    <div class="quote" itemscope itemtype="http://schema.org/CreativeWork">
        <span class="text" itemprop="text">“It is our choices, Harry, that show what we truly are, far more than our abilities.”</span>
        <span>by <small class="author" itemprop="author">J.K. Rowling</small>
        <a href="/author/J-K-Rowling">(about)</a>
        </span>
        <div class="tags">
            Tags:
            <a class="tag" href="/tag/abilities/page/1/">abilities</a>
            <a class="tag" href="/tag/choices/page/1/">choices</a>
        </div>
    </div>
</body>
</html>
'''

if __name__ == '__main__':
    from xpaw import Selector

    selector = Selector(text)

    print('# CSS Selector, content of quotes:')
    for quote in selector.css('div.quote'):
        print(quote.css('span.text')[0].text)

    print('# XPath, content of quotes:')
    for quote in selector.xpath('//div[@class="quote"]'):
        print(quote.xpath('.//span[@class="text"]')[0].text)

    print('# CSS Selector, content of quotes, with HTML tags:')
    for quote in selector.css('div.quote'):
        print(quote.css('span.text')[0].string)

    print('# CSS Selector, quote tags')
    for quote in selector.css('div.quote'):
        print(quote.css('a.tag').text)

    print('# CSS Selector, author urls')
    for quote in selector.css('div.quote'):
        print(quote.css('small+a')[0].attr('href'))



# coding=utf-8

import json
import random

from aiohttp import web
from aiohttp.helpers import BasicAuth
from multidict import MultiDict
from aiohttp import FormData

from xpaw.http import HttpRequest
from xpaw.downloader import Downloader, DownloaderMiddlewareManager
from xpaw.eventbus import EventBus
from xpaw.config import Config
from xpaw import events


async def test_cookies(loop):
    downloader = Downloader(timeout=60, loop=loop)
    seed = str(random.randint(0, 2147483647))
    req = HttpRequest("http://httpbin.org/cookies", cookies={"seed": seed})
    resp = await downloader.download(req)
    assert resp.status == 200
    cookies = json.loads(resp.text)["cookies"]
    assert len(cookies) == 1 and cookies.get("seed") == seed


async def make_auth_server(aiohttp_server, login=None, password=None):
    async def process(request):
        auth_str = request.headers.get('Authorization')
        if auth_str is None:
            if login is not None:
                return web.Response(status=401)
        else:
            auth = BasicAuth.decode(auth_str)
            if login != auth.login or password != auth.password:
                return web.Response(status=401)
        return web.Response(status=200)

    app = web.Application()
    app.router.add_route("GET", "/{tail:.*}", process)
    server = await aiohttp_server(app)
    return server


async def test_basic_auth(aiohttp_server, loop):
    downloader = Downloader(loop=loop)
    server = await make_auth_server(aiohttp_server, login='login', password='pass')

    async def no_auth():
        req = HttpRequest("http://{}:{}/basic-auth/login/pass".format(server.host, server.port))
        resp = await downloader.download(req)
        assert resp.status == 401

    async def str_auth():
        req = HttpRequest("http://{}:{}/basic-auth/login/pass".format(server.host, server.port))
        req.meta['auth'] = 'login:pass'
        resp = await downloader.download(req)
        assert resp.status == 200

    async def tuple_auth():
        req = HttpRequest("http://{}:{}/basic-auth/login/pass".format(server.host, server.port))
        req.meta['auth'] = ('login', 'pass')
        resp = await downloader.download(req)
        assert resp.status == 200

    async def basic_auth():
        req = HttpRequest("http://{}:{}/basic-auth/login/pass".format(server.host, server.port))
        req.meta['auth'] = BasicAuth('login', 'pass')
        resp = await downloader.download(req)
        assert resp.status == 200

    async def url_basic_auth():
        resp = await downloader.download(
            HttpRequest("http://login:pass@{}:{}/basic-auth/login/pass".format(server.host, server.port)))
        assert resp.status == 200

    await no_auth()
    await str_auth()
    await tuple_auth()
    await basic_auth()
    await url_basic_auth()


async def make_params_server(aiohttp_server):
    async def process(request):
        d = {}
        for i in set(request.rel_url.query.keys()):
            v = request.rel_url.query.getall(i)
            if len(v) == 1:
                d[i] = v[0]
            else:
                d[i] = v
        return web.json_response(d, status=200)

    app = web.Application()
    app.router.add_route("GET", "/{tail:.*}", process)
    server = await aiohttp_server(app)
    return server


async def test_params(aiohttp_server, loop):
    server = await make_params_server(aiohttp_server)
    downloader = Downloader(loop=loop)
    args = {'key': 'value', 'none': '', 'list': ['item1', 'item2']}

    async def query_params():
        url = "http://{}:{}/?key=value&none=&list=item1&list=item2".format(server.host, server.port)
        resp = await downloader.download(HttpRequest(url))
        assert args == json.loads(resp.text)

    async def dict_params():
        resp = await downloader.download(HttpRequest("http://{}:{}/".format(server.host, server.port),
                                                     params={'key': 'value', 'none': '', 'list': ['item1', 'item2']}))
        assert args == json.loads(resp.text)

    async def multi_dict_params():
        params = MultiDict()
        params.add('key', 'value')
        params.add('none', '')
        params.add('list', 'item1')
        params.add('list', 'item2')
        resp = await downloader.download(HttpRequest("http://{}:{}/".format(server.host, server.port),
                                                     params=params))
        assert args == json.loads(resp.text)

    await query_params()
    await dict_params()
    await multi_dict_params()


async def make_proxy_auth_server(aiohttp_server, login=None, password=None):
    async def process(request):
        auth_str = request.headers.get('Proxy-Authorization')
        if auth_str is None:
            if login is not None:
                return web.Response(status=401)
        else:
            auth = BasicAuth.decode(auth_str)
            if login != auth.login or password != auth.password:
                return web.Response(status=401)
        return web.Response(body=request.raw_path, status=200)

    app = web.Application()
    app.router.add_route("GET", "/{tail:.*}", process)
    server = await aiohttp_server(app)
    return server


async def test_proxy(aiohttp_server, loop):
    server = await make_proxy_auth_server(aiohttp_server)
    downloader = Downloader(loop=loop)
    seed = str(random.randint(0, 2147483647))
    req = HttpRequest('http://example.com/?seed={}'.format(seed))
    req.meta['proxy'] = '{}:{}'.format(server.host, server.port)
    resp = await downloader.download(req)
    assert resp.text == req.url


async def test_proxy_auth(aiohttp_server, loop):
    downloader = Downloader(loop=loop)
    server = await make_proxy_auth_server(aiohttp_server, login='login', password='pass')

    async def no_auth():
        req = HttpRequest("http://example.com/basic-auth/login/pass")
        req.meta['proxy'] = 'http://{}:{}'.format(server.host, server.port)
        resp = await downloader.download(req)
        assert resp.status == 401

    async def str_auth():
        req = HttpRequest("http://example.com/basic-auth/login/pass")
        req.meta['proxy'] = 'http://{}:{}'.format(server.host, server.port)
        req.meta['proxy_auth'] = 'login:pass'
        resp = await downloader.download(req)
        assert resp.status == 200

    async def tuple_auth():
        req = HttpRequest("http://example.com/basic-auth/login/pass")
        req.meta['proxy'] = 'http://{}:{}'.format(server.host, server.port)
        req.meta['proxy_auth'] = ('login', 'pass')
        resp = await downloader.download(req)
        assert resp.status == 200

    async def basic_auth():
        req = HttpRequest("http://example.com/basic-auth/login/pass")
        req.meta['proxy'] = 'http://{}:{}'.format(server.host, server.port)
        req.meta['proxy_auth'] = BasicAuth('login', 'pass')
        resp = await downloader.download(req)
        assert resp.status == 200

    async def url_basic_auth():
        req = HttpRequest("http://example.com/basic-auth/login/pass")
        req.meta['proxy'] = 'http://login:pass@{}:{}'.format(server.host, server.port)
        resp = await downloader.download(req)
        assert resp.status == 200

    await no_auth()
    await str_auth()
    await tuple_auth()
    await basic_auth()
    await url_basic_auth()


async def make_headers_server(aiohttp_server):
    async def process(request):
        return web.json_response({'headers': dict(request.headers)})

    app = web.Application()
    app.router.add_route("GET", "/{tail:.*}", process)
    server = await aiohttp_server(app)
    return server


async def test_headers(aiohttp_server, loop):
    server = await make_headers_server(aiohttp_server)
    downloader = Downloader(loop=loop)
    headers = {'User-Agent': 'xpaw', 'X-My-Header': 'my-header'}
    resp = await downloader.download(HttpRequest("http://{}:{}".format(server.host, server.port),
                                                 headers=headers))
    assert resp.status == 200
    data = json.loads(resp.body.decode('utf-8'))['headers']
    assert 'User-Agent' in data and data['User-Agent'] == 'xpaw'
    assert 'X-My-Header' in data and data['X-My-Header'] == 'my-header'


async def make_post_server(aiohttp_server):
    async def process(request):
        data = await request.read()
        return web.Response(body=data, status=200)

    async def process_json(request):
        data = await request.json()
        return web.json_response(data, status=200)

    async def process_form(request):
        data = await request.post()
        d = {}
        for i in set(data.keys()):
            v = data.getall(i)
            if len(v) == 1:
                d[i] = v[0]
            else:
                d[i] = v
        return web.json_response(d, status=200)

    app = web.Application()
    app.router.add_route("POST", "/", process)
    app.router.add_route("POST", "/json", process_json)
    app.router.add_route("POST", "/form", process_form)
    server = await aiohttp_server(app)
    return server


async def test_body(aiohttp_server, loop):
    server = await make_post_server(aiohttp_server)
    downloader = Downloader(loop=loop)

    async def post_json():
        json_data = {'key': 'value', 'none': None, 'list': ['item1', 'item2'], 'object': {'name': 'object'}}
        resp = await downloader.download(HttpRequest('http://{}:{}/json'.format(server.host, server.port),
                                                     'POST', body=json_data))
        assert resp.status == 200
        data = json.loads(resp.body.decode())
        assert data == json_data

    async def post_form():
        form_data = {'key': 'value', 'none': '', 'list': ['item1', 'item2']}
        resp = await downloader.download(HttpRequest('http://{}:{}/form'.format(server.host, server.port),
                                                     'POST', body=FormData(form_data)))
        assert resp.status == 200
        data = json.loads(resp.body.decode())
        assert data == form_data

    async def post_str():
        str_data = 'str data: 字符串数据'
        resp = await downloader.download(HttpRequest('http://{}:{}/'.format(server.host, server.port),
                                                     'POST', body=str_data))
        assert resp.status == 200
        body = resp.body.decode()
        assert body == str_data

    async def post_bytes():
        bytes_data = 'bytes data: 字节数据'
        resp = await downloader.download(HttpRequest('http://{}:{}/'.format(server.host, server.port),
                                                     'POST', body=bytes_data.encode()))
        assert resp.status == 200
        body = resp.body.decode()
        assert body == bytes_data

    await post_json()
    await post_form()
    await post_str()
    await post_bytes()


async def make_redirect_server(aiohttp_server):
    async def process(request):
        return web.Response(headers={'Location': 'http://python.org/'}, status=302)

    app = web.Application()
    app.router.add_route("GET", "/{tail:.*}", process)
    server = await aiohttp_server(app)
    return server


async def test_allow_redirects(aiohttp_server, loop):
    server = await make_redirect_server(aiohttp_server)
    downloader = Downloader(loop=loop, allow_redirects=True)
    downloader2 = Downloader(loop=loop, allow_redirects=False)

    resp = await downloader.download(HttpRequest('http://{}:{}/'.format(server.host, server.port)))
    assert resp.status // 100 == 2 and 'python.org' in str(resp.url)

    resp = await downloader2.download(HttpRequest('http://{}:{}/'.format(server.host, server.port)))
    assert resp.status // 100 == 3

    req = HttpRequest('http://{}:{}/'.format(server.host, server.port))
    req.meta['allow_redirects'] = False
    resp = await downloader.download(req)
    assert resp.status // 100 == 3

    req = HttpRequest('http://{}:{}/'.format(server.host, server.port))
    req.meta['allow_redirects'] = True
    resp = await downloader2.download(req)
    assert resp.status // 100 == 2 and 'python.org' in str(resp.url)


class FooDownloadermw:
    def __init__(self, d):
        self.d = d

    def open(self):
        self.d['open'] = ''

    def close(self):
        self.d['close'] = ''

    def handle_request(self, request):
        self.d['handle_request'] = request

    def handle_response(self, request, response):
        self.d['handle_response'] = (request, response)

    def handle_error(self, request, error):
        self.d['handle_error'] = (request, error)


class DummyDownloadermw:
    """
    no method
    """


class FooAsyncDownloaderMw(FooDownloadermw):
    @classmethod
    def from_cluster(cls, cluster):
        return cls(cluster.config['data'])

    async def handle_request(self, request):
        self.d['async_handle_request'] = request

    async def handle_response(self, request, response):
        self.d['async_handle_response'] = (request, response)

    async def handle_error(self, request, error):
        self.d['async_handle_error'] = (request, error)


class Cluster:
    def __init__(self, **kwargs):
        self.event_bus = EventBus()
        self.config = Config(kwargs)


async def test_downloader_middleware_manager_handlers():
    data = {}
    cluster = Cluster(downloader_middlewares=[lambda d=data: FooDownloadermw(d),
                                              DummyDownloadermw,
                                              FooAsyncDownloaderMw],
                      downloader_middlewares_base=None,
                      data=data)
    downloadermw = DownloaderMiddlewareManager.from_cluster(cluster)
    request_obj = object()
    response_obj = object()
    error_obj = object()
    await cluster.event_bus.send(events.cluster_start)
    await downloadermw._handle_request(request_obj)
    await downloadermw._handle_response(request_obj, response_obj)
    await downloadermw._handle_error(request_obj, error_obj)
    await cluster.event_bus.send(events.cluster_shutdown)
    assert 'open' in data and 'close' in data
    assert data['handle_request'] is request_obj
    assert data['handle_response'][0] is request_obj and data['handle_response'][1] is response_obj
    assert data['handle_error'][0] is request_obj and data['handle_error'][1] is error_obj
    assert data['async_handle_request'] is request_obj
    assert data['async_handle_response'][0] is request_obj and data['async_handle_response'][1] is response_obj
    assert data['async_handle_error'][0] is request_obj and data['async_handle_error'][1] is error_obj

    data2 = {}
    cluster2 = Cluster(downloader_middlewares={lambda d=data2: FooDownloadermw(d): 0},
                       downloader_middlewares_base=None,
                       data=data2)
    downloadermw2 = DownloaderMiddlewareManager.from_cluster(cluster2)
    request_obj2 = object()
    await cluster2.event_bus.send(events.cluster_start)
    await downloadermw2._handle_request(request_obj2)
    await cluster2.event_bus.send(events.cluster_shutdown)
    assert 'open' in data2 and 'close' in data2
    assert data2['handle_request'] is request_obj2

    cluster3 = Cluster(downloader_middlewares=None, downloader_middlewares_base=None, data={})
    DownloaderMiddlewareManager.from_cluster(cluster3)



# coding=utf-8

import pytest
from os.path import isfile, isdir, join
from os import remove
import time

from xpaw.cli import main
from xpaw.version import __version__


def test_print_help(capsys):
    with pytest.raises(SystemExit) as excinfo:
        main(argv=['xpaw'])
    assert excinfo.value.code == 0
    out, _ = capsys.readouterr()
    assert out.startswith('usage:')


def test_unknown_command(capsys):
    with pytest.raises(SystemExit) as excinfo:
        main(argv=['xpaw', 'unknown_command'])
    assert excinfo.value.code == 2
    out, _ = capsys.readouterr()
    assert out.startswith('Unknown command: unknown_command')


def test_version(capsys):
    main(argv=['xpaw', 'version'])
    out, _ = capsys.readouterr()
    assert out.strip() == 'xpaw version {}'.format(__version__)


def test_init(tmpdir, capsys):
    proj_name = 'init_test'
    proj_dir = join(str(tmpdir), proj_name)
    main(argv=['xpaw', 'init', proj_dir])
    assert isdir(join(proj_dir))
    assert isfile(join(proj_dir, 'config.py'))
    assert isdir(join(proj_dir, proj_name))
    assert isfile(join(proj_dir, proj_name, '__init__.py'))
    assert isfile(join(proj_dir, proj_name, 'items.py'))
    assert isfile(join(proj_dir, proj_name, 'pipelines.py'))
    assert isfile(join(proj_dir, proj_name, 'spider.py'))
    with pytest.raises(SystemExit) as excinfo:
        main(argv=['xpaw', 'init', proj_dir])
    assert excinfo.value.code == 1
    main(argv=['xpaw', 'crawl', proj_dir, '-s', 'downloader_timeout=0.01', '-l', 'WARNING'])
    _, _ = capsys.readouterr()
    remove(join(proj_dir, 'config.py'))
    with pytest.raises(SystemExit) as excinfo:
        main(argv=['xpaw', 'init', proj_dir])
    assert excinfo.value.code == 1
    _, _ = capsys.readouterr()


def test_init_no_project_dir(capsys):
    with pytest.raises(SystemExit) as excinfo:
        main(argv=['xpaw', 'init'])
    assert excinfo.value.code == 2
    _, _ = capsys.readouterr()


def test_crawl_project(tmpdir, capsys):
    proj_name = 'test_crawl_project'
    proj_dir = join(str(tmpdir), proj_name)
    main(argv=['xpaw', 'init', proj_dir])
    main(argv=['xpaw', 'crawl', proj_dir, '--downloader-timeout=0.01'])
    _, _ = capsys.readouterr()


def test_crawl_no_project_dir(capsys):
    with pytest.raises(SystemExit) as excinfo:
        main(argv=['xpaw', 'crawl'])
    assert excinfo.value.code == 2
    _, _ = capsys.readouterr()


def test_crawl_spider(tmpdir, capsys):
    proj_name = 'test_crawl_spider'
    proj_dir = join(str(tmpdir), proj_name)
    main(argv=['xpaw', 'init', proj_dir])
    with open(join(proj_dir, 'config.py'), 'w') as f:
        f.write('downloader_timeout = 0.01')
    t = time.time()
    main(argv=['xpaw', 'crawl', join(proj_dir, proj_name, 'spider.py'),
               '-c', join(proj_dir, 'config.py')])
    assert time.time() - t < 1
    _, _ = capsys.readouterr()


def test_crawl_spider_no_config_file(tmpdir, capsys):
    proj_name = 'test_crawl_spider'
    proj_dir = join(str(tmpdir), proj_name)
    main(argv=['xpaw', 'init', proj_dir])
    with pytest.raises(SystemExit) as excinfo:
        main(argv=['xpaw', 'crawl', join(proj_dir, proj_name, 'spider.py'),
                   '-c', 'no_such_config.py', '--downloader-timeout=0.01'])
    assert excinfo.value.code == 2
    _, _ = capsys.readouterr()


def test_crawl_no_spider_file(capsys):
    with pytest.raises(SystemExit) as excinfo:
        main(argv=['xpaw', 'crawl', 'dont_exist.py'])
    assert excinfo.value.code == 2
    _, _ = capsys.readouterr()



# coding=utf-8

import pytest

from xpaw.eventbus import EventBus
from xpaw.config import Config
from xpaw.middleware import MiddlewareManager
from xpaw.errors import NotEnabled


class FooMiddleware:
    def __init__(self, d):
        self.d = d

    def open(self):
        self.d['open'] = ''

    def close(self):
        self.d['close'] = ''


class DummyMiddleware:
    """
    no method
    """


class FooEnabledMiddleware:
    def __init__(self, d):
        self.d = d

    def open(self):
        self.d['enabled_open'] = ''

    def close(self):
        self.d['enabled_close'] = ''


class FooDisabledMiddleware:
    def __init__(self, d):
        self.d = d
        raise NotEnabled

    def open(self):
        self.d['disabled_open'] = ''

    def close(self):
        self.d['disabled_close'] = ''


class Cluster:
    def __init__(self, **kwargs):
        self.event_bus = EventBus()
        self.config = Config(kwargs)


async def test_middleware_manager_handlers(monkeypatch):
    @classmethod
    def middleware_list_from_config(cls, config):
        return [lambda data=data: FooMiddleware(data), DummyMiddleware,
                lambda data=data: FooEnabledMiddleware(data), lambda data=data: FooDisabledMiddleware(data)]

    monkeypatch.setattr(MiddlewareManager, '_middleware_list_from_config', middleware_list_from_config)
    data = {}
    middleware_manager = MiddlewareManager.from_cluster(Cluster())
    middleware_manager.open()
    middleware_manager.close()
    assert 'open' in data and 'close' in data
    assert 'enabled_open' in data and 'enabled_close' in data
    assert 'disabled_open' not in data and 'disabled_close' not in data


def test_priority_list_from_config():
    cls = MiddlewareManager._priority_list_from_config
    d = cls('foo', Cluster().config)
    assert d == {}
    with pytest.raises(AssertionError):
        cls('foo', Cluster(foo='a').config)
    d2 = cls('foo', Cluster(foo=['a', 'b', 'c']).config)
    assert d2['a'] == (0, 0) and d2['b'] == (0, 1) and d2['c'] == (0, 2)
    d3 = cls('foo', Cluster(foo={'a': 2, 'b': 1, 'c': 3}).config)
    assert d3['a'] == (2,) and d3['b'] == (1,) and d3['c'] == (3,)


def test_make_component_list():
    cls = MiddlewareManager._make_component_list
    d = cls('foo', Cluster(foo=['a', 'c'], foo_base={'b': 3, 'd': 100}).config)
    assert d == ['a', 'c', 'b', 'd']
    d2 = cls('foo', Cluster(foo=['a', 'b', 'c'], foo_base={'b': 3, 'd': 100}).config)
    assert d2 == ['a', 'b', 'c', 'd']
    d3 = cls('foo', Cluster(foo={'a': 9, 'c': 1}, foo_base={'d': 3, 'b': 4}).config)
    assert d3 == ['c', 'd', 'b', 'a']
    d4 = cls('foo', Cluster(foo={'a': 9, 'c': 1, 'b': 4}, foo_base={'d': 3, 'b': 2}).config)
    assert d4 == ['c', 'd', 'b', 'a']



# coding=utf-8

from os.path import join, exists
import asyncio
import os
import signal
import time
from threading import Thread

from xpaw.spider import Spider
from xpaw.http import HttpRequest
from xpaw.queue import PriorityQueue
from xpaw.run import run_spider
from xpaw.item import Item
from xpaw.errors import IgnoreItem


class SleepSpider(Spider):
    async def start_requests(self):
        await asyncio.sleep(0.2, loop=self.cluster.loop)

    def parse(self, response):
        pass


def test_supervisor():
    run_spider(SleepSpider, downloader_timeout=0.1)


class StartRequestSpider(Spider):
    def start_requests(self):
        yield HttpRequest('http://python.org/')

    async def parse(self, response):
        pass


class BadQueue(PriorityQueue):
    def __init__(self, loop=None, **kwargs):
        super().__init__(loop=loop, **kwargs)
        self.loop = loop

    async def pop(self):
        await super().pop()
        raise RuntimeError('not an error actually')


class BadQueue2(PriorityQueue):
    async def pop(self):
        raise RuntimeError('not an error actually')


def test_coro_terminated():
    run_spider(StartRequestSpider, downloader_clients=2, queue=BadQueue, max_retry_times=0, downloader_timeout=0.1)


def test_coro_terminated2():
    run_spider(StartRequestSpider, downloader_clients=2, queue=BadQueue2, max_retry_times=0, downloader_timeout=0.1)


class ToKillSpider(Spider):
    def start_requests(self):
        yield HttpRequest('http://python.org/')

    async def parse(self, response):
        while True:
            await asyncio.sleep(5, loop=self.cluster.loop)


class ExceptionThread(Thread):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.bucket = []

    def run(self):
        try:
            super().run()
        except Exception as e:
            self.bucket.append(e)
            raise


def test_kill_spider(tmpdir):
    pid_file = join(str(tmpdir), 'pid')
    log_file = join(str(tmpdir), 'log')
    t = ExceptionThread(target=kill_spider, args=(pid_file,))
    t.start()
    run_spider(ToKillSpider, pid_file=pid_file, log_file=log_file)
    t.join()
    assert len(t.bucket) == 0, 'Exception in thread'


def _check_pid(pid):
    try:
        os.kill(pid, 0)
    except OSError:
        return False
    else:
        return True


def kill_spider(pid_file):
    t = 10
    while t > 0 and not exists(pid_file):
        t -= 1
        time.sleep(1)
    assert t > 0
    with open(pid_file, 'rb') as f:
        pid = int(f.read().decode())
    assert _check_pid(pid) is True
    os.kill(pid, signal.SIGTERM)
    t = 10
    while t > 0 and exists(pid_file):
        t -= 1
        time.sleep(1)
    assert t > 0


class FooError(Exception):
    pass


class HandlerDownloaderMiddleware:
    def handle_request(self, request):
        if request.url.endswith('error'):
            raise FooError


class HandlerSpiderMiddleware:
    def handle_input(self, response):
        if response.request.url.endswith('not-found'):
            raise FooError

    def handle_error(self, response, error):
        if isinstance(error, FooError):
            return ()


class HandlerSpider(Spider):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.data = self.config.get('data')
        self.server_address = self.config.get('server_address')

    def start_requests(self):
        yield HttpRequest("http://localhost:80", errback=self.error_back)
        yield HttpRequest("http://localhost:80", dont_filter=True, errback=self.async_error_back)
        yield HttpRequest("http://{}/error".format(self.server_address), errback=self.handle_request_error)
        yield HttpRequest("http://{}/not-found".format(self.server_address), errback=self.handle_input_error)
        yield HttpRequest("http://{}/".format(self.server_address), dont_filter=True)
        yield HttpRequest("http://{}/".format(self.server_address), dont_filter=True, callback=self.generator_parse)
        yield HttpRequest("http://{}/".format(self.server_address), dont_filter=True, callback=self.func_prase)
        yield HttpRequest("http://{}/".format(self.server_address), dont_filter=True, callback=self.async_parse)
        yield HttpRequest("http://{}/".format(self.server_address), dont_filter=True, callback=self.return_list_parse)
        yield HttpRequest("http://{}/".format(self.server_address), dont_filter=True, callback=self.return_none_parse)

    def parse(self, response):
        self.data.add('parse')

    def error_back(self, request, err):
        self.data.add('error_back')
        raise RuntimeError('not an error actually')

    async def async_error_back(self, request, err):
        self.data.add('async_error_back')
        raise RuntimeError('not an error actually')

    def handle_request_error(self, request, error):
        assert isinstance(error, FooError)
        self.data.add('handle_request_error')

    def handle_input_error(self, request, error):
        assert isinstance(error, FooError)
        self.data.add('handle_input_error')

    def generator_parse(self, response):
        self.data.add('generator_parse')
        if response.status / 100 != 2:
            raise RuntimeError('not an error actually')
        # it will never come here
        yield None

    def func_prase(self, response):
        self.data.add('func_parse')
        raise RuntimeError('not an error actually')

    async def async_parse(self, response):
        self.data.add('async_parse')
        raise RuntimeError('not an error actually')

    def return_list_parse(self, response):
        self.data.add('return_list_parse')
        return []

    def return_none_parse(self, response):
        self.data.add('return_none_parse')


def test_spider_handlers():
    data = set()
    run_spider(HandlerSpider, log_level='DEBUG', spider_middlewares=[HandlerSpiderMiddleware],
               downloader_middlewares=[HandlerDownloaderMiddleware], data=data,
               server_address='python.org')
    assert 'parse' in data
    assert 'error_back' in data
    assert 'async_error_back' in data
    assert 'handle_request_error' in data
    assert 'handle_input_error' in data
    assert 'generator_parse' in data
    assert 'func_parse' in data
    assert 'async_parse' in data
    assert 'return_list_parse' in data
    assert 'return_none_parse' in data


class DummyItem(Item):
    pass


class DroppedItem(Item):
    pass


class ErrorItem(Item):
    pass


class FooItemPipeLine:
    def __init__(self, data):
        self.data = data

    @classmethod
    def from_cluster(cls, cluster):
        return cls(cluster.config.get('data'))

    def handle_item(self, item):
        if isinstance(item, DroppedItem):
            raise IgnoreItem
        elif isinstance(item, ErrorItem):
            raise RuntimeError('not an error actually')
        self.data['item'] = item


class ItemSpider(Spider):
    def start_requests(self):
        yield HttpRequest('http://python.org/')

    def parse(self, response):
        yield DroppedItem()
        yield ErrorItem()
        yield DummyItem()


def test_handle_item():
    data = {}
    run_spider(ItemSpider, log_level='DEBUG', data=data, item_pipelines=[FooItemPipeLine])
    assert isinstance(data.get('item'), DummyItem)


class ToDumpSpider(Spider):
    def start_requests(self):
        yield HttpRequest('http://python.org/', callback=self.parse_response, meta={'key': 'value'})

    async def parse_response(self, response):
        while True:
            await asyncio.sleep(5, loop=self.cluster.loop)


class ToLoadSpider(Spider):
    def start_requests(self):
        pass

    async def parse_response(self, response):
        data = self.config.get('data')
        data['url'] = response.request.url
        data['meta'] = response.meta


def test_dump_request(tmpdir):
    dump_dir = str(tmpdir)
    pid_file = join(dump_dir, 'pid')
    t = Thread(target=kill_spider, args=(pid_file,))
    t.start()
    run_spider(ToDumpSpider, dump_dir=dump_dir, pid_file=pid_file, cookie_jar_enabled=True)
    t.join()
    data = {}
    run_spider(ToLoadSpider, dump_dir=dump_dir, data=data, cookie_jar_enabled=True)
    assert data['url'] == 'http://python.org/'
    assert data['meta']['key'] == 'value'



# coding=utf-8

from xpaw.config import BaseConfig, Config, KNOWN_SETTINGS


class TestBaseConfig:
    def test_get(self):
        d = {'key': 'value'}
        config = BaseConfig(values=d)
        assert len(config) == 1
        for k in config:
            assert k == 'key' and config[k] == 'value'
        assert config.get('key') == 'value'
        assert config.get('no_such_key') is None
        assert config.get('no_such_key', 'default') == 'default'
        assert config.get('key', 'default') == 'value'
        assert config['key'] == 'value'
        assert config['no_such_key'] is None
        assert ('key' in config) is True
        assert ('no_such_key' in config) is False

    def test_get_bool(self):
        d = {'bool_true': 'true', 'bool_True': 'True', 'bool_false': 'false', 'bool_False': 'False',
             'bool_int1': '1', 'bool_int0': '0', 'bool_none': '...',
             'true': True, 'false': False}
        config = BaseConfig(values=d)
        assert config.getbool('true') is True
        assert config.getbool('false') is False
        assert config.getbool('bool_true') is True
        assert config.getbool('bool_True') is True
        assert config.getbool('bool_false') is False
        assert config.getbool('bool_False') is False
        assert config.getbool('bool_int1') is True
        assert config.getbool('bool_int0') is False
        assert config.getbool('bool_none') is None
        assert config.getbool('bool_no') is None
        assert config.getbool('bool_no', True) is True
        assert config.getbool('bool_no', False) is False
        assert config.getbool('bool_no', '...') is None
        assert config.getbool('bool_true', False) is True
        assert config.getbool('bool_false', True) is False

    def test_get_int(self):
        d = {'int_1': 1, 'int_str_1': '1', 'int_none': '...'}
        config = BaseConfig(values=d)
        assert config.getint('int_1') == 1
        assert config.getint('int_str_1') == 1
        assert config.getint('int_none') is None
        assert config.getint('int_no') is None
        assert config.getint('int_1', 0) == 1
        assert config.getint('int_none', 0) is None
        assert config.getint('int_no', 0) == 0

    def test_get_float(self):
        d = {'float_1.1': 1.1, 'float_str_1.1': '1.1', 'float_none': '...'}
        config = BaseConfig(values=d)
        assert config.getfloat('float_1.1') == 1.1
        assert config.getfloat('float_str_1.1') == 1.1
        assert config.getfloat('float_none') is None
        assert config.getfloat('float_no') is None
        assert config.getfloat('float_1.1', 0) == 1.1
        assert config.getfloat('float_none', 0) is None
        assert config.getfloat('float_no', 0) == 0

    def test_get_list(self):
        d = {'list': [1, 2], 'tuple': (1, 2), 'single': 1, 'list_str': '1,2'}
        config = BaseConfig(values=d)
        assert config.getlist('list') == [1, 2]
        assert config.getlist('tuple') == [1, 2]
        assert config.getlist('single') == [1]
        assert config.getlist('list_str') == ['1', '2']
        assert config.getlist('list', [1]) == [1, 2]
        assert config.getlist('no_such_list') is None
        assert config.getlist('no_such_list', [1]) == [1]

    def test_set(self):
        config = BaseConfig()
        config.set('key', 'value')
        assert len(config) == 1 and config['key'] == 'value'

        config.set('key', 'value2')
        assert config['key'] == 'value2'

        config.set('key2', 'value')
        assert len(config) == 2 and config['key2'] == 'value'

    def test_set_item(self):
        config = BaseConfig()
        config['key'] = 'value'
        assert len(config) == 1 and config['key'] == 'value'

        config['key'] = 'value2'
        assert config['key'] == 'value2'

        config['key2'] = 'value'
        assert len(config) == 2 and config['key2'] == 'value'

    def test_update(self):
        config = BaseConfig()
        config.update({'key': 'value'})
        assert len(config) == 1 and config['key'] == 'value'

        config.update({'key': 'value2', 'key2': 'value'})
        assert len(config) == 2 and config['key'] == 'value2' and config['key2'] == 'value'

    def test_update_by_base_config(self):
        c1 = BaseConfig()
        c1.set('k1', 'c1_k1')
        c2 = BaseConfig({'k1': 'c2_k1', 'k2': 'c2_k2'})
        c2.set('k3', 'c2_k3')
        c1.update(c2)
        assert len(c1) == 3
        assert c1['k1'] == 'c2_k1'
        assert c1['k2'] == 'c2_k2'
        assert c1['k3'] == 'c2_k3'

    def test_copy(self):
        c1 = BaseConfig()
        c1.set('dict', {'k': 'v'})
        c2 = c1.copy()
        c1['dict']['k'] = 'vv'
        assert c2['dict']['k'] == 'v'

    def test_delete(self):
        c = BaseConfig()
        c.set('k1', 'v1')
        c.set('k2', 'v2')
        c.set('k3', 'v3')
        c.delete('k2')
        assert len(c) == 2 and 'k2' not in c
        c.delete('k1')
        assert len(c) == 1 and 'k1' not in c
        del c['k3']
        assert len(c) == 0 and 'k3' not in c


def test_config():
    c = Config()
    assert len(c) > 0
    for k in c:
        assert k in KNOWN_SETTINGS and c.get(k) == KNOWN_SETTINGS[k].value



# coding=utf-8

from collections import MutableMapping

from . import config


class BaseItem:
    """
    Super class for all items.
    """


class Field(dict):
    """
    Field consist of metadata
    """


class Item(MutableMapping, BaseItem):
    def __init__(self, **kwargs):
        self.fields = {}
        for k in dir(self):
            v = getattr(self, k)
            if isinstance(v, Field):
                self.fields[k] = v
        self.values = {}
        for k, v in kwargs.items():
            self[k] = v

    def __getitem__(self, key):
        if key not in self:
            return None
        v = self.values[key]
        t = self.fields[key].get('type')
        if t:
            if isinstance(t, str):
                if t == 'str':
                    v = str(v)
                elif t == 'int':
                    v = config.getint(v)
                elif t == 'float':
                    v = config.getfloat(v)
                elif t == 'bool':
                    v = config.getbool(v)
                else:
                    raise ValueError('Unsupported item filed type: {}'.format(t))
            else:
                v = t(v)
        return v

    def __contains__(self, name):
        return name in self.values

    def __setitem__(self, key, value):
        if key in self.fields:
            self.values[key] = value
        else:
            raise KeyError(key)

    def __delitem__(self, key):
        del self.values[key]

    def __len__(self):
        return len(self.values)

    def __iter__(self):
        return iter(self.values)

    def keys(self):
        return self.values.keys()

    def __repr__(self):
        return repr(dict(self))

    def copy(self):
        return self.__class__(**self.values)



# coding=utf-8

from urllib.parse import urljoin
import json

from xpaw import Spider, HttpRequest, Selector, run_spider


class QuotesSpider(Spider):
    quotes = []

    def start_requests(self):
        yield HttpRequest('http://quotes.toscrape.com/', callback=self.parse)

    def parse(self, response):
        selector = Selector(response.text)
        for quote in selector.css('div.quote'):
            text = quote.css('span.text')[0].text
            author = quote.css('small.author')[0].text
            author_url = quote.css('small+a')[0].attr('href')
            author_url = urljoin(str(response.url), author_url)
            tags = quote.css('div.tags a').text
            self.quotes.append(dict(text=text, tags=tags,
                                    author=author, author_url=author_url))
        next_page = selector.css('li.next a')
        if len(next_page) > 0:
            next_page_url = urljoin(str(response.url), next_page[0].attr('href'))
            yield HttpRequest(next_page_url, callback=self.parse)

    def close(self):
        with open('quotes.json', 'w') as f:
            json.dump(self.quotes, f, ensure_ascii=False, indent=4)


if __name__ == '__main__':
    run_spider(QuotesSpider, log_level='DEBUG')



# coding=utf-8

from urllib.parse import urljoin

from xpaw import Spider, HttpRequest, Selector, run_spider


class XPathQuotesSpider(Spider):
    def start_requests(self):
        yield HttpRequest('http://quotes.toscrape.com/', callback=self.parse)

    def parse(self, response):
        selector = Selector(response.text)
        for quote in selector.xpath('//div[@class="quote"]'):
            text = quote.xpath('.//span[@itemprop="text"]')[0].text
            author = quote.xpath('.//small[@itemprop="author"]')[0].text
            author_url = quote.xpath('.//span/a/@href')[0].text
            author_url = urljoin(str(response.url), author_url)
            tags = quote.xpath('.//div[@class="tags"]/a').text
            self.log('quote: %s', dict(text=text, tags=tags,
                                       author=author, author_url=author_url))
        next_page_url = selector.xpath('//li[@class="next"]/a/@href')[0].text
        self.log('next page url: %s', next_page_url)


if __name__ == '__main__':
    run_spider(XPathQuotesSpider, log_level='DEBUG')



# coding=utf-8

from xpaw.stats import StatsCollector, DummyStatsCollector


class TestStatsCollector:
    def test_get_none_key(self):
        stats = StatsCollector()
        assert stats.get('key') is None
        assert stats.get('key', 'default') is 'default'
        stats.set('key', 'value')
        assert stats.get('key', 'default') == 'value'

    def test_set_value(self):
        stats = StatsCollector()
        stats.set('key1', 0)
        assert stats.get('key1') == 0
        stats.set_max('key1', 2)
        assert stats.get('key1') == 2
        stats.set_max('key1', 1)
        assert stats.get('key1') == 2
        stats.set_max('key2', 3)
        assert stats.get('key2') == 3
        stats.set('key3', 5)
        assert stats.get('key3') == 5
        stats.set_min('key3', 4)
        assert stats.get('key3') == 4
        stats.set_min('key3', 6)
        assert stats.get('key3') == 4
        stats.set_min('key4', 1)
        assert stats.get('key4') == 1
        stats.set_default('key5', 6)
        assert stats.get('key5') == 6
        stats.set_default('key5', 5)
        assert stats.get('key5') == 6

    def test_inc_value(self):
        stats = StatsCollector()
        stats.set('key1', 1)
        stats.inc('key1')
        assert stats.get('key1') == 2
        stats.inc('key1', 2)
        assert stats.get('key1') == 4
        stats.inc('key1', 5, start=0)
        assert stats.get('key1') == 9
        stats.inc('key2')
        assert stats.get('key2') == 1
        stats.inc('key3', start=1)
        assert stats.get('key3') == 2
        stats.inc('key4', 2, start=3)
        assert stats.get('key4') == 5

    def test_clear_stats(self):
        stats = StatsCollector()
        stats.set('key1', 1)
        stats.set('key2', 2)
        assert stats.stats == {'key1': 1, 'key2': 2}
        stats.clear()
        assert stats.stats == {}

    def test_set_stats(self):
        stats = StatsCollector()
        stats.set_stats({'key1': 1, 'key2': 2})
        assert stats.stats == {'key1': 1, 'key2': 2}

    def test_remove_key(self):
        stats = StatsCollector()
        stats.set('key', 'value')
        assert stats.get('key') == 'value'
        stats.remove('key')
        assert stats.get('key') is None
        assert stats.get('key', 'default') == 'default'


class TestDummyStatsCollector:
    def test_set_value(self):
        stats = DummyStatsCollector()
        assert stats.get('key') is None
        stats.set('key', 'value')
        assert stats.get('key') is None
        assert stats.get('key', 'default') == 'default'
        stats.set_min('key', 0)
        assert stats.get('key') is None
        stats.set_max('key', 0)
        assert stats.get('key') is None
        stats.set_default('key', 'value')
        assert stats.get('key') is None

    def test_inc_value(self):
        stats = DummyStatsCollector()
        stats.inc('key')
        assert stats.get('key') is None

    def test_set_stats(self):
        stats = DummyStatsCollector()
        stats.set_stats({'key1': 1, 'key2': 2})
        assert stats.stats == {}



# coding=utf-8

import logging

from .http import HttpRequest

log = logging.getLogger(__name__)


class DepthMiddleware:
    def __init__(self, max_depth):
        self._max_depth = max_depth

    def __repr__(self):
        cls_name = self.__class__.__name__
        return '{}(max_depth={})'.format(cls_name, repr(self._max_depth))

    @classmethod
    def from_cluster(cls, cluster):
        config = cluster.config
        max_depth = config.getint("max_depth")
        return cls(max_depth=max_depth)

    def handle_output(self, response, result):
        depth = response.meta.get("depth", 0) + 1
        for r in result:
            if isinstance(r, HttpRequest):
                r.meta["depth"] = depth
                if self._max_depth is None or depth <= self._max_depth:
                    yield r
                else:
                    log.debug("Abort %s: depth > %s", r, self._max_depth)
            else:
                yield r

    def handle_start_requests(self, result):
        for r in result:
            if isinstance(r, HttpRequest):
                r.meta['depth'] = 0
            yield r



# coding=utf-8

import logging
from os.path import join, isfile

from . import utils
from . import events

log = logging.getLogger(__name__)


class HashDupeFilter:
    def __init__(self, dump_dir=None):
        self._dump_dir = dump_dir
        self._hash = set()

    @classmethod
    def from_cluster(cls, cluster):
        config = cluster.config
        dupe_filter = cls(dump_dir=utils.get_dump_dir(config))
        cluster.event_bus.subscribe(dupe_filter.open, events.cluster_start)
        cluster.event_bus.subscribe(dupe_filter.close, events.cluster_shutdown)
        return dupe_filter

    def is_duplicated(self, request):
        if request.dont_filter:
            return False
        h = utils.request_fingerprint(request)
        if h in self._hash:
            log.debug("%s is duplicated", request)
            return True
        self._hash.add(h)
        return False

    def clear(self):
        self._hash.clear()

    def open(self):
        if self._dump_dir:
            file = join(self._dump_dir, 'dupe_filter')
            if isfile(file):
                with open(file, 'r') as f:
                    for h in f:
                        self._hash.add(h.rstrip())

    def close(self):
        if self._dump_dir:
            with open(join(self._dump_dir, 'dupe_filter'), 'w') as f:
                for h in self._hash:
                    f.write(h + '\n')



# coding=utf-8

from urllib.parse import urljoin

from xpaw import Spider, HttpRequest, Selector

from .items import QuotesItem


class QuotesSpider(Spider):
    def start_requests(self):
        yield HttpRequest('http://quotes.toscrape.com/', callback=self.parse)

    def parse(self, response):
        selector = Selector(response.text)
        for quote in selector.css('div.quote'):
            text = quote.css('span.text')[0].text
            author = quote.css('small.author')[0].text
            author_url = quote.css('small+a')[0].attr('href')
            author_url = urljoin(str(response.url), author_url)
            tags = quote.css('div.tags a').text
            yield QuotesItem(text=text, tags=tags,
                             author=author, author_url=author_url)
        next_page = selector.css('li.next a')
        if len(next_page) > 0:
            next_page_url = urljoin(str(response.url), next_page[0].attr('href'))
            yield HttpRequest(next_page_url, callback=self.parse)



# coding=utf-8

import os
import re
import cgi
import sys
import hashlib
import logging
from importlib import import_module
import string
from os.path import isfile, exists
import inspect

from yarl import URL

PY35 = sys.version_info >= (3, 5)
PY36 = sys.version_info >= (3, 6)


def load_object(path):
    if isinstance(path, str):
        dot = path.rindex(".")
        module, name = path[:dot], path[dot + 1:]
        mod = import_module(module)
        return getattr(mod, name)
    return path


def configure_logger(name, config):
    log_level = config.get('log_level').upper()
    log_format = config.get('log_format')
    log_dateformat = config.get('log_dateformat')
    logger = logging.getLogger(name)
    logger.setLevel(log_level)
    log_file = config.get('log_file')
    if log_file:
        handler = logging.FileHandler(log_file)
    else:
        handler = logging.StreamHandler()
    handler.setLevel(log_level)
    formatter = logging.Formatter(log_format, log_dateformat)
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    return handler


def remove_logger(name):
    logging.getLogger(name).handlers.clear()


def get_encoding_from_header(content_type):
    if content_type:
        content_type, params = cgi.parse_header(content_type)
        if "charset" in params:
            return params["charset"]


_charset_flag = re.compile(r"""<meta.*?charset=["']*(.+?)["'>]""", flags=re.I)
_pragma_flag = re.compile(r"""<meta.*?content=["']*;?charset=(.+?)["'>]""", flags=re.I)
_xml_flag = re.compile(r"""^<\?xml.*?encoding=["']*(.+?)["'>]""")


def get_encoding_from_content(content):
    if isinstance(content, bytes):
        content = content.decode("ascii", errors="ignore")
    elif not isinstance(content, str):
        raise ValueError("content should be bytes or str")
    s = _charset_flag.search(content)
    if s:
        return s.group(1).strip()
    s = _pragma_flag.search(content)
    if s:
        return s.group(1).strip()
    s = _xml_flag.search(content)
    if s:
        return s.group(1).strip()


def request_fingerprint(request):
    sha1 = hashlib.sha1()
    sha1.update(to_bytes(request.method))
    if isinstance(request.url, str):
        url = URL(request.url)
    else:
        url = request.url
    queries = []
    for k, v in url.query.items():
        queries.append('{}={}'.format(k, v))
    if request.params:
        for k, v in request.params.items():
            queries.append('{}={}'.format(k, v))
    queries.sort()
    sha1.update(to_bytes('{}://{}{}:{}?{}'.format(url.scheme, url.host, url.path, url.port, '&'.join(queries))))
    sha1.update(request.body or b'')
    return sha1.hexdigest()


def to_bytes(data, encoding=None):
    if isinstance(data, bytes):
        return data
    if isinstance(data, str):
        return data.encode(encoding or "utf-8")
    raise TypeError("Need bytes or str, got {}".format(type(data).__name__))


def render_template_file(path, **kwargs):
    if path.endswith(".tmpl"):
        with open(path, "rb") as f:
            raw = f.read().decode("utf-8")
        content = string.Template(raw).substitute(**kwargs)
        render_path = path[:-len(".tmpl")]
        with open(render_path, "wb") as f:
            f.write(content.encode("utf-8"))
        os.remove(path)


_camelcase_invalid_chars = re.compile('[^a-zA-Z\d]')


def string_camelcase(s):
    return _camelcase_invalid_chars.sub('', s.title())


class AsyncGenWrapper:
    def __init__(self, gen):
        if not gen:
            gen = ()
        self.iter = gen.__iter__()

    async def __aiter__(self):
        return self

    async def __anext__(self):
        try:
            return next(self.iter)
        except StopIteration:
            raise StopAsyncIteration


async def iterable_to_list(gen):
    res = []
    if not hasattr(gen, '__aiter__'):
        gen = AsyncGenWrapper(gen)
    async for r in gen:
        res.append(r)
    return res


def cmp(a, b):
    return (a > b) - (a < b)


def be_daemon():
    if os.fork():
        os._exit(0)
    os.setsid()
    if os.fork():
        os._exit(0)
    os.umask(0o22)
    os.closerange(0, 3)
    fd_null = os.open(os.devnull, os.O_RDWR)
    if fd_null != 0:
        os.dup2(fd_null, 0)
    os.dup2(fd_null, 1)
    os.dup2(fd_null, 2)


def load_config(fname):
    if fname is None or not isfile(fname):
        raise ValueError('{} is not a file'.format(fname))
    code = compile(open(fname, 'rb').read(), fname, 'exec')
    cfg = {
        "__builtins__": __builtins__,
        "__name__": "__config__",
        "__file__": fname,
        "__doc__": None,
        "__package__": None
    }
    exec(code, cfg, cfg)
    return cfg


def iter_settings(config):
    for key, value in config.items():
        if not key.startswith('_') and not inspect.ismodule(value) and not inspect.isfunction(value):
            yield key, value


def get_dump_dir(config):
    dump_dir = config.get('dump_dir')
    if dump_dir:
        if not exists(dump_dir):
            os.makedirs(dump_dir, 755)
        return dump_dir


def request_to_dict(request):
    callback = request.callback
    if inspect.ismethod(callback):
        callback = callback.__name__
    errback = request.errback
    if inspect.ismethod(errback):
        errback = errback.__name__
    meta = dict(request.meta)
    if 'cookie_jar' in meta:
        del meta['cookie_jar']
    d = {
        'url': request.url,
        'method': request.method,
        'body': request.body,
        'params': request.params,
        'headers': request.headers,
        'cookies': request.cookies,
        'meta': meta,
        'priority': request.priority,
        'dont_filter': request.dont_filter,
        'callback': callback,
        'errback': errback,
        '_class': request.__module__ + '.' + request.__class__.__name__
    }
    return d


def request_from_dict(d):
    req_cls = load_object(d.pop('_class'))
    return req_cls(**d)



# coding=utf-8

import re
import json

import pytest

from xpaw.config import Config
from xpaw.http import HttpRequest, HttpResponse
from xpaw.downloadermws import *
from xpaw.errors import IgnoreRequest, ClientError, NotEnabled, TimeoutError
from xpaw.version import __version__
from xpaw.downloader import Downloader
from xpaw.eventbus import EventBus


class Cluster:
    def __init__(self, loop=None, **kwargs):
        self.loop = loop
        self.config = Config(kwargs)
        self.event_bus = EventBus()


class TestImitatingProxyMiddleware:
    def test_handle_request(self):
        mw = ImitatingProxyMiddleware.from_cluster(Cluster(imitating_proxy_enabled=True))
        req = HttpRequest("http://example.com")
        mw.handle_request(req)
        assert re.search(r"^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}", req.headers["X-Forwarded-For"])
        assert req.headers['Via'] == '{} xpaw'.format(__version__)

    def test_not_enabled(self):
        with pytest.raises(NotEnabled):
            ImitatingProxyMiddleware.from_cluster(Cluster())


class TestDefaultHeadersMiddleware:
    def test_handle_request(self):
        default_headers = {"User-Agent": "xpaw", "Connection": "keep-alive"}
        req_headers = {"User-Agent": "xpaw-test", "Connection": "keep-alive"}
        mw = DefaultHeadersMiddleware.from_cluster(Cluster(default_headers=default_headers))
        req = HttpRequest("http://example.com", headers={"User-Agent": "xpaw-test"})
        mw.handle_request(req)
        assert req_headers == req.headers

    def test_not_enabled(self, loop):
        with pytest.raises(NotEnabled):
            DefaultHeadersMiddleware.from_cluster(Cluster(loop=loop, default_headers=None))


class Random:
    def __init__(self):
        self.iter = 0

    def choice(self, seq):
        res = seq[self.iter % len(seq)]
        self.iter += 1
        return res


class TestProxyMiddleware:
    def test_proxy_str(self, loop):
        proxy = '153.10.32.18:3128'
        mw = ProxyMiddleware.from_cluster(Cluster(proxy=proxy, loop=loop))
        req = HttpRequest(URL("http://example.com"))
        mw.handle_request(req)
        assert req.meta['proxy'] == '153.10.32.18:3128'

    def test_proxy_list(self, monkeypatch, loop):
        monkeypatch.setattr(random, 'choice', Random().choice)
        proxy_list = ["105.12.103.232:3128", "16.82.3.20:3128"]
        mw = ProxyMiddleware.from_cluster(Cluster(proxy=proxy_list, loop=loop))
        for i in range(len(proxy_list)):
            req = HttpRequest("http://example.com")
            mw.handle_request(req)
            assert req.meta['proxy'] == proxy_list[i]

        req2 = HttpRequest('ftp://example.com')
        mw.handle_request(req2)
        assert 'proxy' not in req2.meta

    def test_proxy_dict(self, monkeypatch, loop):
        monkeypatch.setattr(random, 'choice', Random().choice)
        proxy_dict = {'http': ['132.39.13.100:3128', '18.39.9.10:3128'], 'https': '177.13.233.2:3128'}
        mw = ProxyMiddleware.from_cluster(Cluster(proxy=proxy_dict, loop=loop))
        req_list = []
        for i in ['http://example.com', 'https://example.com', 'http://example.com', 'http://example.com']:
            req_list.append(HttpRequest(i))
        res = ['132.39.13.100:3128', '177.13.233.2:3128', '132.39.13.100:3128', '18.39.9.10:3128']
        for i in range(len(req_list)):
            mw.handle_request(req_list[i])
            assert req_list[i].meta['proxy'] == res[i]

    async def test_not_enabled(self, loop):
        with pytest.raises(NotEnabled):
            ProxyMiddleware.from_cluster(Cluster(loop=loop))


class TestRetryMiddleware:
    def test_handle_reponse(self):
        mw = RetryMiddleware.from_cluster(Cluster(retry_http_status=(500,)))
        req = HttpRequest("http://example.com")
        resp = HttpResponse(URL("http://example.com"), 502)
        assert mw.handle_response(req, resp) is None
        req2 = HttpRequest("http://example.com")
        resp2 = HttpResponse(URL("http://example.com"), 500)
        retry_req2 = mw.handle_response(req2, resp2)
        assert retry_req2.meta['retry_times'] == 1
        assert str(retry_req2.url) == str(req2.url)
        req3 = HttpRequest("http://example.com")
        resp3 = HttpResponse(URL("http://example.com"), 500)
        req3.meta['retry_times'] = 2
        retry_req3 = mw.handle_response(req3, resp3)
        assert retry_req3.meta['retry_times'] == 3
        assert str(retry_req3.url) == str(req3.url)
        req4 = HttpRequest("http://example.com")
        req4.meta['retry_times'] = 3
        resp4 = HttpResponse(URL("http://example.com"), 500)
        with pytest.raises(IgnoreRequest):
            mw.handle_response(req4, resp4)

    def test_handle_error(self):
        mw = RetryMiddleware.from_cluster(Cluster())
        req = HttpRequest("http://example.com")
        err = ValueError()
        assert mw.handle_error(req, err) is None
        for err in [ClientError(), TimeoutError()]:
            retry_req = mw.handle_error(req, err)
            assert isinstance(retry_req, HttpRequest) and str(retry_req.url) == str(req.url)

    def test_retry(self):
        max_retry_times = 2
        mw = RetryMiddleware.from_cluster(Cluster(max_retry_times=max_retry_times,
                                                  retry_http_status=(500,)))
        req = HttpRequest("http://example.com")
        for i in range(max_retry_times):
            retry_req = mw.retry(req, "")
            assert isinstance(retry_req, HttpRequest) and str(retry_req.url) == str(req.url)
            req = retry_req
        with pytest.raises(IgnoreRequest):
            mw.retry(req, "")

    def test_not_enabled(self):
        with pytest.raises(NotEnabled):
            RetryMiddleware.from_cluster(Cluster(retry_enabled=False))

    def test_match_status(self):
        assert RetryMiddleware.match_status("200", 200) is True
        assert RetryMiddleware.match_status(200, 200) is True
        assert RetryMiddleware.match_status("2xX", 201) is True
        assert RetryMiddleware.match_status("40x", 403) is True
        assert RetryMiddleware.match_status("40X", 403) is True
        assert RetryMiddleware.match_status("50x", 403) is False
        assert RetryMiddleware.match_status("~20X", 200) is False
        assert RetryMiddleware.match_status("!20x", 400) is True
        assert RetryMiddleware.match_status("0200", 200) is False


class TestCookiesMiddleware:
    def test_not_enabled(self):
        with pytest.raises(NotEnabled):
            CookiesMiddleware.from_cluster(Cluster())

    async def test_cookie_jar(self, loop):
        mw = CookiesMiddleware.from_cluster(Cluster(cookie_jar_enabled=True, loop=loop))
        downloader = Downloader(timeout=60, loop=loop)
        seed = str(random.randint(0, 2147483647))
        req = HttpRequest("http://httpbin.org/cookies/set?seed={}".format(seed))
        mw.handle_request(req)
        await downloader.download(req)
        req2 = HttpRequest("http://httpbin.org/cookies")
        mw.handle_request(req2)
        resp = await downloader.download(req2)
        assert resp.status == 200
        cookies = json.loads(resp.text)["cookies"]
        assert len(cookies) == 1 and cookies.get("seed") == seed
        req3 = HttpRequest("http://httpbin.org/cookies/delete?seed=")
        mw.handle_request(req3)
        await downloader.download(req3)
        req4 = HttpRequest("http://httpbin.org/cookies")
        mw.handle_request(req4)
        resp = await downloader.download(req4)
        assert resp.status == 200
        cookies = json.loads(resp.text)["cookies"]
        assert len(cookies) == 0

    async def test_dump(self, loop, tmpdir):
        mw = CookiesMiddleware.from_cluster(Cluster(cookie_jar_enabled=True, loop=loop, dump_dir=str(tmpdir)))
        downloader = Downloader(timeout=60, loop=loop)
        seed = str(random.randint(0, 2147483647))
        req = HttpRequest("http://httpbin.org/cookies/set?seed={}".format(seed))
        mw.handle_request(req)
        await downloader.download(req)
        mw.close()
        mw2 = CookiesMiddleware.from_cluster(Cluster(cookie_jar_enabled=True, loop=loop, dump_dir=str(tmpdir)))
        mw2.open()
        req2 = HttpRequest("http://httpbin.org/cookies")
        mw2.handle_request(req2)
        resp = await downloader.download(req2)
        assert resp.status == 200
        cookies = json.loads(resp.text)["cookies"]
        assert len(cookies) == 1 and cookies.get("seed") == seed


class TestSpeedLimitMiddleware:
    async def test_value_error(self, loop):
        with pytest.raises(ValueError):
            SpeedLimitMiddleware.from_cluster(Cluster(speed_limit_enabled=True,
                                                      speed_limit_rate=0,
                                                      speed_limit_burst=1,
                                                      loop=loop))
        with pytest.raises(ValueError):
            SpeedLimitMiddleware.from_cluster(Cluster(speed_limit_enabled=True,
                                                      speed_limit_rate=1,
                                                      speed_limit_burst=0,
                                                      loop=loop))

    async def test_not_enabled(self, loop):
        with pytest.raises(NotEnabled):
            SpeedLimitMiddleware.from_cluster(Cluster(speed_limit_rate=1,
                                                      speed_limit_burst=1,
                                                      loop=loop))

    async def test_handle_request(self, loop):
        class Counter:
            def __init__(self):
                self.n = 0

            def inc(self):
                self.n += 1

        async def processor():
            while True:
                await mw.handle_request(None)
                counter.inc()

        counter = Counter()
        mw = SpeedLimitMiddleware.from_cluster(Cluster(speed_limit_enabled=True,
                                                       speed_limit_rate=1000,
                                                       speed_limit_burst=5,
                                                       loop=loop))
        futures = []
        for i in range(100):
            futures.append(asyncio.ensure_future(processor(), loop=loop))
        mw.open()
        await asyncio.sleep(0.1, loop=loop)
        mw.close()
        for f in futures:
            assert f.cancel() is True
        await asyncio.sleep(0.01, loop=loop)
        assert counter.n <= 105


class TestUserAgentMiddleware:
    def test_not_enabled(self):
        with pytest.raises(NotEnabled):
            UserAgentMiddleware.from_cluster(Cluster(user_agent=None))

    def test_static_user_agent(self):
        user_agent = 'test user agent'
        mw = UserAgentMiddleware.from_cluster(Cluster(user_agent=user_agent))
        req = HttpRequest('http://example.com')
        mw.handle_request(req)
        assert req.headers.get('User-Agent') == user_agent

    def test_gen_user_agent(self):
        mw = UserAgentMiddleware.from_cluster(Cluster(user_agent=':desktop,chrome'))
        req = HttpRequest('http://example.com')
        mw.handle_request(req)
        assert 'Chrome' in req.headers.get('User-Agent')

        mw2 = UserAgentMiddleware.from_cluster(Cluster(user_agent=':mobile,chrome'))
        req2 = HttpRequest('http://example.com')
        mw2.handle_request(req2)
        assert 'CriOS' in req2.headers.get('User-Agent') and 'Mobile' in req2.headers.get('User-Agent')

    def test_unknown_user_agent_description(self):
        with pytest.raises(ValueError):
            UserAgentMiddleware.from_cluster(Cluster(user_agent=':unknown'))

    def test_random_user_agent(self):
        mw = UserAgentMiddleware.from_cluster(Cluster(random_user_agent=True))
        req = HttpRequest('http://example.com')
        req2 = HttpRequest('http://example.com')
        mw.handle_request(req)
        mw.handle_request(req2)
        assert 'User-Agent' in req.headers
        assert req.headers.get('User-Agent') != req2.headers.get('User-Agent')
        assert 'Chrome' in req.headers.get('User-Agent')

    def test_random_user_agent2(self):
        mw = UserAgentMiddleware.from_cluster(Cluster(user_agent=':mobile', random_user_agent=True))
        for i in range(30):
            req = HttpRequest('http://example.com')
            mw.handle_request(req)
            assert 'User-Agent' in req.headers
            assert 'CriOS' in req.headers.get('User-Agent') and 'Mobile' in req.headers.get('User-Agent')



# coding=utf-8

import pytest

from xpaw.eventbus import EventBus

event1 = object()
event2 = object()


class FooClass:
    def __init__(self):
        self.count1 = 0
        self.count2 = 0
        self.value = 0

    def method(self):
        """
        method
        """

    @staticmethod
    def staticmethod():
        """
        static method
        """

    def method1(self):
        self.count1 += 1

    async def method2(self):
        self.count2 += 1

    def method_set_value(self, value):
        self.value = value

    def method_raise_error(self):
        raise RuntimeError('not an error actually')


def test_raise_value_error():
    def func():
        """
        function
        """

    eventbus = EventBus()
    with pytest.raises(ValueError):
        eventbus.subscribe(func, event1)
    with pytest.raises(ValueError):
        eventbus.subscribe(FooClass().staticmethod, event1)
    eventbus.subscribe(FooClass().method, event1)


async def test_subscribe():
    eventbus = EventBus()
    obj = FooClass()
    eventbus.subscribe(obj.method1, event1)
    eventbus.subscribe(obj.method2, event2)
    await eventbus.send(event1)
    assert obj.count1 == 1 and obj.count2 == 0
    await eventbus.send(event2)
    assert obj.count1 == 1 and obj.count2 == 1


async def test_subscribe_multi_times():
    eventbus = EventBus()
    obj = FooClass()
    eventbus.subscribe(obj.method1, event1)
    eventbus.subscribe(obj.method1, event1)
    await eventbus.send(event1)
    assert obj.count1 == 1


async def test_unsubscribe():
    eventbus = EventBus()
    obj = FooClass()
    eventbus.subscribe(obj.method1, event1)
    await eventbus.send(event1)
    assert obj.count1 == 1
    eventbus.unsubscribe(obj.method1, event1)
    await eventbus.send(event1)
    assert obj.count1 == 1


async def test_unsubscribe_multi_times():
    eventbus = EventBus()
    obj = FooClass()
    eventbus.subscribe(obj.method1, event1)
    eventbus.unsubscribe(obj.method1, event1)
    eventbus.unsubscribe(obj.method1, event1)


async def test_send_with_parameters():
    eventbus = EventBus()
    obj = FooClass()
    eventbus.subscribe(obj.method_set_value, event1)
    await eventbus.send(event1, value=-1)
    assert obj.value == -1


async def test_send_unknown_event():
    eventbus = EventBus()
    obj = FooClass()
    await eventbus.send(event1)
    eventbus.subscribe(obj.method1, event1)
    await eventbus.send(event2)
    assert obj.count1 == 0


async def test_unknown_unsubscribe():
    eventbus = EventBus()
    obj = FooClass()
    eventbus.subscribe(obj.method1, event1)
    eventbus.unsubscribe(obj.method1, event2)
    eventbus.unsubscribe(obj.method2, event1)
    eventbus.unsubscribe(obj.method, event2)
    await eventbus.send(event1)
    assert obj.count1 == 1


async def test_raise_error_in_callback():
    eventbus = EventBus()
    obj = FooClass()
    eventbus.subscribe(obj.method_raise_error, event1)
    await eventbus.send(event1)


async def test_del_weak_ref():
    eventbus = EventBus()
    obj1 = FooClass()
    obj2 = FooClass()
    eventbus.subscribe(obj1.method1, event1)
    eventbus.subscribe(obj2.method1, event1)
    await eventbus.send(event1)
    assert obj1.count1 == 1 and obj2.count1 == 1
    del obj2
    assert len(eventbus._refs[event1]) == 2
    await eventbus.send(event1)
    assert len(eventbus._refs[event1]) == 1
    assert obj1.count1 == 2



# coding=utf-8

from lxml import etree, cssselect

_text_type_config = {
    'html': {
        'parser_cls': etree.HTMLParser,
        'tostring_method': 'html',
        'css_translator': cssselect.LxmlHTMLTranslator()
    },
    'xml': {
        'parser_cls': etree.XMLParser,
        'tostring_method': 'xml',
        'css_translator': cssselect.LxmlTranslator()
    }
}


def _get_text_type(text_type):
    if text_type is None:
        return 'html'
    if text_type in _text_type_config:
        return text_type
    raise ValueError('Invalid document type: {}'.format(text_type))


def create_root_node(text, parser_cls):
    return etree.fromstring(text, parser=parser_cls())


class Selector:
    def __init__(self, text=None, root=None, text_type=None):
        self.type = _get_text_type(text_type)
        c = _text_type_config[self.type]
        self._parser_cls = c['parser_cls']
        self._tostring_method = c['tostring_method']
        self._css_translator = c['css_translator']
        if text is not None:
            if not isinstance(text, str):
                raise TypeError("'text' argument must be str")
            root = create_root_node(text, parser_cls=self._parser_cls)
        elif root is None:
            raise ValueError("Needs either text or root argument")
        self.root = root

    def xpath(self, xpath, **kwargs):
        try:
            res = self.root.xpath(xpath, smart_strings=False, **kwargs)
        except Exception:
            return SelectorList([])
        if not isinstance(res, list):
            res = [res]
        return SelectorList([self.__class__(root=i) for i in res])

    def css(self, css, **kwargs):
        try:
            path = self._css_translator.css_to_xpath(css)
            xpath = etree.XPath(path, smart_strings=False, **kwargs)
            res = xpath(self.root)
        except Exception:
            return SelectorList([])
        if not isinstance(res, list):
            res = [res]
        return SelectorList([self.__class__(root=i) for i in res])

    @property
    def string(self):
        try:
            return etree.tostring(self.root, encoding="unicode", method=self._tostring_method, with_tail=False)
        except TypeError:
            return str(self.root)

    @property
    def text(self):
        try:
            return etree.tostring(self.root, encoding="unicode", method="text", with_tail=False)
        except TypeError:
            return str(self.root)

    def attr(self, name):
        res = self.xpath('@' + name)
        if len(res) > 0:
            return res[0].text


class SelectorList(list):
    def __getitem__(self, item):
        obj = super().__getitem__(item)
        return self.__class__(obj) if isinstance(item, slice) else obj

    def xpath(self, xpath, **kwargs):
        res = self.__class__()
        for i in self:
            res += i.xpath(xpath, **kwargs)
        return res

    def css(self, css, **kwargs):
        res = self.__class__()
        for i in self:
            res += i.css(css, **kwargs)
        return res

    @property
    def string(self):
        return [i.string for i in self]

    @property
    def text(self):
        return [i.text for i in self]

    def attr(self, name):
        return [i.attr(name) for i in self]



# coding=utf-8

import logging

from .middleware import MiddlewareManager

log = logging.getLogger(__name__)


class ExtensionManager(MiddlewareManager):
    @classmethod
    def _middleware_list_from_config(cls, config):
        return cls._make_component_list('extensions', config)



# coding=utf-8

from .utils import get_encoding_from_header, get_encoding_from_content


class HttpRequest:
    def __init__(self, url, method="GET", body=None, params=None, headers=None, cookies=None,
                 meta=None, priority=None, dont_filter=False, callback=None, errback=None):
        """
        Construct an HTTP request.
        """
        self.url = url
        self.method = method
        self.body = body
        self.params = params
        self.headers = headers or {}
        self.cookies = cookies or {}
        self._meta = dict(meta) if meta else {}
        self.priority = priority
        self.dont_filter = dont_filter
        self.callback = callback
        self.errback = errback

    def __str__(self):
        return '<{}, {}>'.format(self.method, self.url)

    __repr__ = __str__

    @property
    def meta(self):
        return self._meta

    def copy(self):
        return self.replace()

    def replace(self, **kwargs):
        for i in ["url", "method", "body", "params", "headers", "cookies",
                  "meta", "priority", "dont_filter", "callback", "errback"]:
            kwargs.setdefault(i, getattr(self, i))
        return type(self)(**kwargs)


class HttpResponse:
    def __init__(self, url, status, body=None, headers=None, cookies=None,
                 request=None):
        """
        Construct an HTTP response.
        """
        self.url = url
        self.status = int(status)
        self.body = body
        self.headers = headers or {}
        self.cookies = cookies or {}
        self.request = request

    def __str__(self):
        return '<{}, {}>'.format(self.status, self.url)

    __repr__ = __str__

    @property
    def encoding(self):
        if hasattr(self, "_encoding"):
            return self._encoding
        encoding = get_encoding_from_header(self.headers.get("Content-Type"))
        if not encoding and self.body:
            encoding = get_encoding_from_content(self.body)
        self._encoding = encoding or "utf-8"
        return self._encoding

    @encoding.setter
    def encoding(self, value):
        self._encoding = value

    @property
    def text(self):
        if hasattr(self, "_text") and self._text:
            return self._text
        if not self.body:
            return ""
        self._text = self.body.decode(self.encoding, errors="replace")
        return self._text

    @property
    def meta(self):
        if self.request:
            return self.request.meta

    def copy(self):
        return self.replace()

    def replace(self, **kwargs):
        for i in ["url", "status", "body", "headers", "cookies", "request"]:
            kwargs.setdefault(i, getattr(self, i))
        return type(self)(**kwargs)



# coding=utf-8

import pytest

from xpaw.item import Item, Field


class FooItem(Item):
    f1 = Field()
    f2 = Field()


def test_item():
    with pytest.raises(KeyError):
        FooItem(f3='v3')
    item = FooItem(f1='v1')
    assert len(item) == 1 and 'f1' in item
    assert item['f1'] == 'v1' and item['f2'] is None
    item['f2'] = 'v2'
    assert item.keys() == {'f1', 'f2'}
    s = {}
    for k in item:
        s[k] = item[k]
    assert s == {'f1': 'v1', 'f2': 'v2'}
    assert len(item) == 2 and 'f2' in item
    assert item['f2'] == 'v2'
    item_copy = item.copy()
    del item['f2']
    assert len(item) == 1 and 'f2' not in item
    assert len(item_copy) == 2 and item_copy['f1'] == 'v1' and item_copy['f2'] == 'v2'
    item['f2'] = 'new_v2'
    assert item['f2'] == 'new_v2'
    repr_str = repr(item)
    assert repr_str == "{'f1': 'v1', 'f2': 'new_v2'}" or repr_str == "{'f2': 'new_v2', 'f1': 'v1'}"


class FieldTypeItem(Item):
    none_field = Field()
    str_field = Field(type='str')
    int_field = Field(type='int')
    float_field = Field(type='float')
    bool_field = Field(type='bool')
    func_field = Field(type=int)
    error_field = Field(type='error type')


def test_field_type():
    item = FieldTypeItem(str_field=1, int_field='1', float_field='1', bool_field='1', func_field='1', error_field='1')
    assert item['none_field'] is None
    assert item['str_field'] == '1'
    assert isinstance(item['int_field'], int) and item['int_field'] == 1
    assert isinstance(item['float_field'], float) and item['float_field'] == 1
    assert item['bool_field'] is True
    assert isinstance(item['func_field'], int) and item['func_field'] == 1
    with pytest.raises(ValueError):
        print(item['error_field'])



# coding=utf-8

import time
import logging
import asyncio
from asyncio import Semaphore
from collections import deque
from heapq import heappush, heappop
from os.path import join, isfile
import pickle

from . import utils
from . import events

log = logging.getLogger(__name__)


class FifoQueue:
    def __init__(self, dump_dir=None, loop=None):
        self._queue = deque()
        self._dump_dir = dump_dir
        self._loop = loop or asyncio.get_event_loop()
        self._semaphore = Semaphore(0, loop=self._loop)

    def __len__(self):
        return len(self._queue)

    @classmethod
    def from_cluster(cls, cluster):
        queue = cls(dump_dir=utils.get_dump_dir(cluster.config), loop=cluster.loop)
        cluster.event_bus.subscribe(queue.open, events.cluster_start)
        cluster.event_bus.subscribe(queue.close, events.cluster_shutdown)
        return queue

    async def push(self, request):
        self._queue.append(request)
        self._semaphore.release()

    async def pop(self):
        await self._semaphore.acquire()
        return self._queue.popleft()

    async def open(self):
        if self._dump_dir:
            file = join(self._dump_dir, 'queue')
            if isfile(file):
                with open(file, 'rb') as f:
                    arr = pickle.load(f)
                for a in arr:
                    r = utils.request_from_dict(a)
                    await self.push(r)

    async def close(self):
        if self._dump_dir:
            reqs = []
            while len(self) > 0:
                r = await self.pop()
                reqs.append(utils.request_to_dict(r))
            with open(join(self._dump_dir, 'queue'), 'wb') as f:
                pickle.dump(reqs, f)


class LifoQueue(FifoQueue):
    async def pop(self):
        await self._semaphore.acquire()
        return self._queue.pop()

    async def open(self):
        if self._dump_dir:
            file = join(self._dump_dir, 'queue')
            if isfile(file):
                with open(file, 'rb') as f:
                    arr = pickle.load(f)
                arr.reverse()
                for a in arr:
                    r = utils.request_from_dict(a)
                    await self.push(r)


class PriorityQueue:
    def __init__(self, dump_dir=None, loop=None):
        self._queue = []
        self._dump_dir = dump_dir
        self._loop = loop or asyncio.get_event_loop()
        self._semaphore = Semaphore(0, loop=self._loop)

    def __len__(self):
        return len(self._queue)

    @classmethod
    def from_cluster(cls, cluster):
        queue = cls(dump_dir=utils.get_dump_dir(cluster.config), loop=cluster.loop)
        cluster.event_bus.subscribe(queue.open, events.cluster_start)
        cluster.event_bus.subscribe(queue.close, events.cluster_shutdown)
        return queue

    async def push(self, request):
        heappush(self._queue, _PriorityQueueItem(request))
        self._semaphore.release()

    async def pop(self):
        await self._semaphore.acquire()
        item = heappop(self._queue)
        return item.request

    async def open(self):
        if self._dump_dir:
            file = join(self._dump_dir, 'queue')
            if isfile(file):
                with open(file, 'rb') as f:
                    arr = pickle.load(f)
                for a in arr:
                    r = utils.request_from_dict(a)
                    await self.push(r)

    async def close(self):
        if self._dump_dir:
            reqs = []
            while len(self) > 0:
                r = await self.pop()
                reqs.append(utils.request_to_dict(r))
            with open(join(self._dump_dir, 'queue'), 'wb') as f:
                pickle.dump(reqs, f)


class _PriorityQueueItem:
    def __init__(self, request):
        self.request = request
        self.priority = self.request.priority or 0
        self.now = time.time()

    def __cmp__(self, other):
        return utils.cmp((-self.priority, self.now), (-other.priority, other.now))

    def __lt__(self, other):
        return self.__cmp__(other) < 0



# coding=utf-8

from xpaw.http import HttpRequest
from xpaw.dupefilter import HashDupeFilter


def run_any_dupe_filter(f):
    r_get = HttpRequest("http://example.com")
    r_get_port_80 = HttpRequest("http://example.com:80")
    r_get_port_81 = HttpRequest("http://example.com:81")
    r_get_dont_filter = HttpRequest("http://example.com", dont_filter=True)
    r_get_dir = HttpRequest("http://example.com/")
    r_get_post = HttpRequest("http://example.com/post")
    r_post = HttpRequest("http://example.com/post", "POST")
    r_post_dir = HttpRequest("http://example.com/post/", "POST")
    r_post_data = HttpRequest("http://example.com/post", "POST", body=b'data')
    r_get_param = HttpRequest("http://example.com/get", params={'k1': 'v1'})
    r_get_query = HttpRequest("http://example.com/get?k1=v1")
    r_get_param_2 = HttpRequest("http://example.com/get", params={'k1': 'v1', 'k2': 'v2'})
    r_get_query_2 = HttpRequest("http://example.com/get?k2=v2&k1=v1")
    r_get_query_param = HttpRequest("http://example.com/get?k1=v1", params={'k2': 'v2'})
    assert f.is_duplicated(r_get) is False
    assert f.is_duplicated(r_get_port_80) is True
    assert f.is_duplicated(r_get_port_81) is False
    assert f.is_duplicated(r_get) is True
    assert f.is_duplicated(r_get_dont_filter) is False
    assert f.is_duplicated(r_get_dir) is True
    assert f.is_duplicated(r_get_post) is False
    assert f.is_duplicated(r_post) is False
    assert f.is_duplicated(r_post_dir) is False
    assert f.is_duplicated(r_post_data) is False
    assert f.is_duplicated(r_get_param) is False
    assert f.is_duplicated(r_get_query) is True
    assert f.is_duplicated(r_get_param_2) is False
    assert f.is_duplicated(r_get_query_2) is True
    assert f.is_duplicated(r_get_query_param) is True


class TestHashDupeFilter:
    def test_is_duplicated(self):
        run_any_dupe_filter(HashDupeFilter())

    def test_clear(self):
        f = HashDupeFilter()
        r_get = HttpRequest("http://example.com")
        assert f.is_duplicated(r_get) is False
        assert f.is_duplicated(r_get) is True
        f.clear()
        assert f.is_duplicated(r_get) is False

    def test_dump(self, tmpdir):
        f = HashDupeFilter(dump_dir=str(tmpdir))
        r_get = HttpRequest("http://example.com")
        assert f.is_duplicated(r_get) is False
        assert f.is_duplicated(r_get) is True
        f.close()
        f2 = HashDupeFilter(dump_dir=str(tmpdir))
        f2.open()
        assert f2.is_duplicated(r_get) is True



# coding=utf-8

import time


def log_time(name):
    def wrapper(func):
        def run(*args, **kwargs):
            start = time.time()
            res = func(*args, **kwargs)
            print('The execution time of {}: {}'.format(name, time.time() - start))
            return res

        return run

    return wrapper



# coding=utf-8

import random
from queue import PriorityQueue
from heapq import heappush, heappop

from benchmarks.utils import log_time


class ListHeapPriorityQueue:
    def __init__(self):
        self.q = []

    def push(self, item):
        heappush(self.q, item)

    def pop(self):
        return heappop(self.q)


@log_time('prepare benchmark data')
def prepare_benchmark_data(push_rate=0.5, total=0):
    data = []
    n = 0
    while len(data) + n < total:
        r = random.random()
        if n == 0 or r < push_rate:
            data.append(('push', random.randint(0, 2147483647)))
            n += 1
        else:
            data.append(('pop', None))
            n -= 1
    while n > 0:
        data.append(('pop', None))
        n -= 1
    return data


@log_time('system priority queue')
def benchmark_system_priority_queue(data):
    q = PriorityQueue()
    for d in data:
        op, v = d
        if op == 'push':
            q.put_nowait(v)
        else:
            q.get_nowait()


@log_time('list heap priority queue')
def benchmark_list_heap_priority_queue(data):
    q = ListHeapPriorityQueue()
    for d in data:
        op, v = d
        if op == 'push':
            q.push(v)
        else:
            q.pop()


def main():
    print('--------------------------------')
    print('push rate: 0.6    total: 1000000')
    print('--------------------------------')
    data1 = prepare_benchmark_data(push_rate=0.6, total=1000000)
    benchmark_system_priority_queue(data1)
    benchmark_list_heap_priority_queue(data1)
    print('--------------------------------')
    print('push rate: 0.8    total: 1000000')
    print('--------------------------------')
    data2 = prepare_benchmark_data(push_rate=0.8, total=1000000)
    benchmark_system_priority_queue(data2)
    benchmark_list_heap_priority_queue(data2)


if __name__ == '__main__':
    main()



# coding=utf-8

import json
from os.path import dirname, join

home_dir = dirname(dirname(__file__))


class QuotesPipeline:
    def __init__(self):
        self.data = []

    def handle_item(self, item):
        self.data.append(dict(item))

    def close(self):
        with open(join(home_dir, 'quotes.json'), 'w') as f:
            json.dump(self.data, f, ensure_ascii=False, indent=4)



# coding=utf-8

from xpaw import Spider, HttpRequest, Selector, run_spider


class AsyncGeneratorSpider(Spider):
    """
    Need Python 3.6+
    """

    async def start_requests(self):
        yield HttpRequest("http://quotes.toscrape.com/", callback=self.parse)

    async def parse(self, response):
        selector = Selector(response.text)
        tags = selector.xpath("//div[contains(@class, 'tags-box')]//a").text
        self.log("Top ten tags: %s", tags)
        yield HttpRequest("http://quotes.toscrape.com/", callback=self.parse)


if __name__ == '__main__':
    run_spider(AsyncGeneratorSpider, log_level="DEBUG")



# coding=utf-8


class StatsCollector:
    def __init__(self):
        self._stats = {}

    def get(self, key, default=None):
        return self._stats.get(key, default)

    def set(self, key, value):
        self._stats[key] = value

    def set_default(self, key, default=None):
        self._stats.setdefault(key, default)

    def set_min(self, key, value):
        self._stats[key] = min(self._stats.setdefault(key, value), value)

    def set_max(self, key, value):
        self._stats[key] = max(self._stats.setdefault(key, value), value)

    def inc(self, key, value=1, start=0):
        self._stats[key] = self._stats.setdefault(key, start) + value

    def clear(self):
        self._stats.clear()

    def remove(self, key):
        if key in self._stats:
            del self._stats[key]

    @property
    def stats(self):
        return self._stats

    def set_stats(self, stats):
        self._stats = stats


class DummyStatsCollector(StatsCollector):
    def get(self, key, default=None):
        return default

    def set(self, key, value):
        pass

    def set_default(self, key, default=None):
        pass

    def set_min(self, key, value):
        pass

    def set_max(self, key, value):
        pass

    def inc(self, key, value=1, start=0):
        pass

    def set_stats(self, stats):
        pass



# coding=utf-8

# Add patch to avoid 'TIME_WAIT'
from . import _patch

from aiohttp import FormData

from .http import HttpRequest, HttpResponse
from .downloader import Downloader
from .spider import Spider, every
from .selector import Selector
from .item import Item, Field
from .run import run_spider, run_crawler, make_requests

__all__ = ['FormData',
           'HttpRequest', 'HttpResponse',
           'Downloader',
           'Spider', 'every',
           'Selector',
           'Item', 'Field',
           'run_spider', 'run_crawler', 'make_requests']

del _patch



# coding=utf-8

from xpaw.pipeline import ItemPipelineManager
from xpaw.config import Config
from xpaw.eventbus import EventBus
from xpaw import events


class FooItemPipeline:
    def __init__(self, d):
        self.d = d

    def handle_item(self, item):
        self.d['handle_item'] = item

    def open(self):
        self.d['open'] = ''

    def close(self):
        self.d['close'] = ''


class DummyItemPipeline:
    """
    no method
    """


class FooAsyncItemPipeline(FooItemPipeline):
    @classmethod
    def from_cluster(cls, cluster):
        return cls(cluster.config['data'])

    async def handle_item(self, item):
        self.d['async_handle_item'] = item


class Cluster:
    def __init__(self, **kwargs):
        self.event_bus = EventBus()
        self.config = Config(kwargs)


async def test_item_pipeline_manager():
    data = {}
    cluster = Cluster(item_pipelines=[lambda d=data: FooItemPipeline(d),
                                      DummyItemPipeline,
                                      FooAsyncItemPipeline],
                      item_pipelines_base=None,
                      data=data)
    pipeline = ItemPipelineManager.from_cluster(cluster)
    obj = object()
    await cluster.event_bus.send(events.cluster_start)
    await pipeline.handle_item(obj)
    await cluster.event_bus.send(events.cluster_shutdown)
    assert 'open' in data and 'close' in data
    assert data['handle_item'] is obj and data['async_handle_item'] is obj

    data2 = {}
    cluster2 = Cluster(item_pipelines={lambda d=data2: FooItemPipeline(d): 0},
                       item_pipelines_base=None,
                       data=data2)
    pipeline2 = ItemPipelineManager.from_cluster(cluster2)
    obj2 = object()
    await cluster2.event_bus.send(events.cluster_start)
    await pipeline2.handle_item(obj2)
    await cluster2.event_bus.send(events.cluster_shutdown)
    assert 'open' in data2 and 'close' in data2
    assert data2['handle_item'] is obj2

    cluster3 = Cluster(item_pipelines=None, item_pipelines_base=None, data={})
    ItemPipelineManager.from_cluster(cluster3)



# coding=utf-8

import asyncio
import logging
from asyncio import CancelledError
import time
import inspect

from .downloader import Downloader
from .http import HttpRequest, HttpResponse
from .errors import IgnoreRequest, IgnoreItem, StopCluster
from .downloader import DownloaderMiddlewareManager
from .spider import Spider, SpiderMiddlewareManager
from .eventbus import EventBus
from . import events
from .extension import ExtensionManager
from .item import BaseItem
from .pipeline import ItemPipelineManager
from . import utils

log = logging.getLogger(__name__)


class LocalCluster:
    def __init__(self, config, loop=None):
        self.config = config
        self.loop = loop or asyncio.get_event_loop()
        self.event_bus = EventBus()
        self.stats_collector = self._new_object_from_cluster(self.config.get('stats_collector'), self)
        self.queue = self._new_object_from_cluster(self.config.get('queue'), self)
        self.dupe_filter = self._new_object_from_cluster(self.config.get('dupe_filter'), self)
        self.downloader = Downloader(timeout=self.config.getfloat('downloader_timeout'),
                                     verify_ssl=self.config.getbool('verify_ssl'),
                                     allow_redirects=self.config.getbool('allow_redirects'),
                                     loop=self.loop)
        self.spider = self._new_object_from_cluster(self.config.get('spider'), self)
        assert isinstance(self.spider, Spider), 'spider must inherit from the Spider class'
        log.info('Spider: %s', str(self.spider))
        self.downloadermw = DownloaderMiddlewareManager.from_cluster(self)
        log.info('Downloader middlewares: %s', self._log_objects(self.downloadermw.components))
        self.spidermw = SpiderMiddlewareManager.from_cluster(self)
        log.info('Spider middlewares: %s', self._log_objects(self.spidermw.components))
        self.item_pipeline = ItemPipelineManager.from_cluster(self)
        log.info('Item pipelines: %s', self._log_objects(self.item_pipeline.components))
        self.extension = ExtensionManager.from_cluster(self)
        log.info('Extensions: %s', self._log_objects(self.extension.components))
        self._job_futures = None
        self._job_futures_done = None
        self._req_in_job = None
        self._start_future = None
        self._supervisor_future = None
        self._is_running = False
        self._run_lock = None

    async def run(self):
        if self._is_running:
            return
        self._is_running = True
        await self._run()
        self._run_lock = asyncio.Future(loop=self.loop)
        await self._run_lock
        self._run_lock = None

    async def _run(self):
        await self.event_bus.send(events.cluster_start)
        self._supervisor_future = asyncio.ensure_future(self._supervisor(), loop=self.loop)
        self._start_future = asyncio.ensure_future(self._push_start_requests(), loop=self.loop)
        downloader_clients = self.config.getint("downloader_clients")
        log.info("The maximum number of simultaneous clients: %s", downloader_clients)
        self._job_futures = []
        for i in range(downloader_clients):
            f = asyncio.ensure_future(self._pull_requests(i), loop=self.loop)
            self._job_futures.append(f)
        self._job_futures_done = set()
        self._req_in_job = [None] * downloader_clients
        log.info('Cluster is loaded')

    def stop(self):
        if not self._is_running:
            return
        self._is_running = False
        asyncio.ensure_future(self._shutdown(), loop=self.loop)

    async def _shutdown(self):
        log.info("Shutdown now")
        cancelled_futures = []
        if self._job_futures:
            for f in self._job_futures:
                f.cancel()
                cancelled_futures.append(f)
            self._job_futures = None
            self._job_futures_done = None
        if self._start_future:
            self._start_future.cancel()
            cancelled_futures.append(self._start_future)
            self._start_future = None
        if self._supervisor_future:
            self._supervisor_future.cancel()
            cancelled_futures.append(self._supervisor_future)
            self._supervisor_future = None
        # put back the unfinished requests
        if self._req_in_job:
            for r in self._req_in_job:
                if r:
                    await self.queue.push(r)
            self._req_in_job = None
        await self.event_bus.send(events.cluster_shutdown)
        # wait cancelled futures
        await asyncio.wait(cancelled_futures, loop=self.loop)
        log.info('Cluster is unloaded')
        if self._run_lock:
            self._run_lock.set_result(True)

    async def _supervisor(self):
        timeout = self.config.getfloat("downloader_timeout")
        while True:
            await asyncio.sleep(timeout, loop=self.loop)
            for i in range(len(self._job_futures)):
                f = self._job_futures[i]
                if f.done():
                    if i not in self._job_futures_done:
                        self._job_futures_done.add(i)
                        reason = "This future is cancelled" if f.cancelled() else str(f.exception())
                        log.error("Worker[%s] is shutdown: %s", i, reason)
                        self._req_in_job[i] = None
            if self._all_done():
                break
        self.stop()

    def _all_done(self):
        if self._start_future.done() and len(self.queue) <= 0:
            no_active = True
            for i in range(len(self._job_futures)):
                if self._req_in_job[i]:
                    no_active = False
                    break
            return no_active
        elif len(self._job_futures_done) == len(self._job_futures):
            log.error('No alive worker')
            return True
        return False

    async def _push_start_requests(self):
        try:
            if hasattr(self.spider.start_requests, "cron_job"):
                tick = self.spider.start_requests.cron_tick
            else:
                tick = 0
            while True:
                t = time.time()
                res = await self.spidermw.start_requests(self.spider)
                for r in res:
                    if isinstance(r, HttpRequest):
                        await self.schedule(r)
                if tick <= 0:
                    break
                t = time.time() - t
                if t < tick:
                    await asyncio.sleep(tick - t, loop=self.loop)
        except CancelledError:
            raise
        except Exception:
            log.warning("Failed to generate start requests", exc_info=True)

    async def _pull_requests(self, coro_id):
        while True:
            req = await self.queue.pop()
            log.debug("%s -> worker[%s]", req, coro_id)
            self._req_in_job[coro_id] = req
            try:
                resp = await self.downloadermw.download(self.downloader, req)
            except CancelledError:
                raise
            except Exception as e:
                if isinstance(e, IgnoreRequest):
                    await self.event_bus.send(events.request_ignored, request=req, error=e)
                await self.spider.request_error(req, e)
            else:
                await self._handle_response(resp)
            self._req_in_job[coro_id] = None
            # check if it's all done
            if self._all_done():
                self.stop()

    async def _handle_response(self, resp):
        if isinstance(resp, HttpRequest):
            await self.schedule(resp)
        elif isinstance(resp, HttpResponse):
            await self.event_bus.send(events.response_received, response=resp)
            try:
                result = await self.spidermw.parse(self.spider, resp)
                for r in result:
                    await self._handle_result(r)
            except CancelledError:
                raise
            except Exception as e:
                if isinstance(e, StopCluster):
                    log.info('Request to stop cluster: %s', e)
                    self.stop()
                else:
                    log.warning("Failed to parse %s", resp, exc_info=True)

    async def _handle_result(self, result):
        if isinstance(result, HttpRequest):
            await self.schedule(result)
        elif isinstance(result, (BaseItem, dict)):
            try:
                await self.item_pipeline.handle_item(result)
            except CancelledError:
                raise
            except Exception as e:
                if isinstance(e, IgnoreItem):
                    await self.event_bus.send(events.item_ignored, item=result, error=e)
                else:
                    log.warning("Failed to handle %s", result, exc_info=True)
            else:
                await self.event_bus.send(events.item_scraped, item=result)

    @staticmethod
    def _new_object_from_cluster(cls_path, cluster):
        obj_cls = utils.load_object(cls_path)
        if hasattr(obj_cls, "from_cluster"):
            obj = obj_cls.from_cluster(cluster)
        else:
            obj = obj_cls()
        return obj

    async def schedule(self, request):
        res = self.dupe_filter.is_duplicated(request)
        if inspect.iscoroutine(res):
            res = await res
        if not res:
            await self.event_bus.send(events.request_scheduled, request=request)
            await self.queue.push(request)

    @staticmethod
    def _log_objects(objects):
        if objects:
            return ''.join(['\n\t' + str(i) for i in objects])
        return ''



# coding=utf-8


from xpaw.extension import ExtensionManager
from xpaw.config import Config
from xpaw.eventbus import EventBus
from xpaw import events


class FooExtension:
    def __init__(self, d):
        self.d = d

    def open(self):
        self.d['open'] = ''

    def close(self):
        self.d['close'] = ''


class DummyExtension:
    """
    no method
    """


class Cluster:
    def __init__(self, **kwargs):
        self.event_bus = EventBus()
        self.config = Config(kwargs)


async def test_extension_manager():
    data = {}
    cluster = Cluster(extensions=[lambda d=data: FooExtension(d),
                                  DummyExtension],
                      extensions_base=None,
                      data=data)
    extension = ExtensionManager.from_cluster(cluster)
    await cluster.event_bus.send(events.cluster_start)
    await cluster.event_bus.send(events.cluster_shutdown)
    assert 'open' in data and 'close' in data

    data2 = {}
    cluster2 = Cluster(extensions={lambda d=data2: FooExtension(d): 0},
                       extensions_base=None,
                       data=data2)
    extension2 = ExtensionManager.from_cluster(cluster2)
    await cluster2.event_bus.send(events.cluster_start)
    await cluster2.event_bus.send(events.cluster_shutdown)
    assert 'open' in data2 and 'close' in data2

    cluster3 = Cluster(extensions=None, extensions_base=None, data={})
    ExtensionManager.from_cluster(cluster3)



# coding=utf-8

from xpaw import Spider, HttpRequest, Selector, every, run_spider


class CronJobSpider(Spider):
    @every(seconds=10)
    def start_requests(self):
        yield HttpRequest("http://news.qq.com/", callback=self.parse, dont_filter=True)

    def parse(self, response):
        selector = Selector(response.text)
        major_news = selector.css("div.major a.linkto").text
        self.log("Major news:")
        for i in range(len(major_news)):
            self.log("%s: %s", i + 1, major_news[i])


if __name__ == '__main__':
    run_spider(CronJobSpider, log_level='DEBUG')



# coding=utf-8

from os.path import join

import pytest

from xpaw.spider import Spider
from xpaw.cli import main
from xpaw.run import run_crawler, run_spider, make_requests
from xpaw.http import HttpRequest, HttpResponse
from xpaw.errors import IgnoreRequest


def test_run_crawler(tmpdir):
    proj_name = 'test_run_crawler'
    proj_dir = join(str(tmpdir), proj_name)
    main(argv=['xpaw', 'init', proj_dir])
    run_crawler(proj_dir, downloader_timeout=0.01, log_level='DEBUG')


def test_run_crawler_bad_config(tmpdir, capsys):
    proj_dir = join(str(tmpdir))
    config_file = join(proj_dir, 'config.py')
    with open(config_file, 'w') as f:
        f.write('bad config')
    with pytest.raises(RuntimeError):
        run_crawler(proj_dir, log_level='DEBUG')
    _, _ = capsys.readouterr()


def test_failed_to_create_cluster(tmpdir, capsys):
    proj_dir = join(str(tmpdir))
    with pytest.raises(Exception):
        run_crawler(proj_dir, log_level='DEBUG')
    _, _ = capsys.readouterr()


class DummySpider(Spider):
    def start_requests(self):
        pass

    def parse(self, response):
        pass


def test_run_spider():
    run_spider(DummySpider, downloader_timeout=0.1, log_level='DEBUG')


def test_make_requests():
    requests = [None, 'http://localhost:80', 'http://python.org/', HttpRequest('http://python.org')]
    results = make_requests(requests)
    assert len(results) == len(requests)
    assert results[0] is None
    assert isinstance(results[1], IgnoreRequest)
    assert isinstance(results[2], HttpResponse) and results[2].status == 200
    assert isinstance(results[3], HttpResponse) and results[3].status == 200



# coding=utf-8

import pytest
from aiohttp.helpers import parse_mimetype

from xpaw.selector import Selector


class TestXPathSelector:
    def test_selector_list(self):
        html = """<li>a</li><div><ul><li>b</li><li>c</li></ul></div><ul><li>d</li></ul>"""
        s = Selector(html)
        assert s.xpath("//li/text()")[0].text == "a"
        assert s.xpath("//li/text()")[-1].text == "d"
        assert s.xpath("//div").xpath(".//li").text[0] == "b"
        assert s.xpath("//div").xpath(".//li").text[-1] == "c"

    def test_attribute_selection(self):
        html = """<a href="http://example.com/" target=_blank>"""
        s = Selector(html)
        assert s.xpath('//a')[0].xpath('@href')[0].text == 'http://example.com/'
        assert s.xpath("//a/@href")[0].text == "http://example.com/"
        assert s.xpath("//a/@target")[0].text == "_blank"

    def test_text_selection(self):
        html = """<div><p>expression: <var>x</var>+<var>y</var>=<var>z</var></p></div>"""
        s = Selector(html)
        assert s.xpath("//var/text()")[0].text == "x"
        assert s.xpath("//var")[0].text == "x"
        assert s.xpath("//var[last()]").text == ["z"]
        assert s.xpath("//var/text()").text == ["x", "y", "z"]
        assert s.xpath("//var").text == ["x", "y", "z"]
        assert s.xpath("//p")[0].text == "expression: x+y=z"
        assert s.xpath("//p/text()")[0].string == "expression: "

    def test_diff_between_string_and_text(self):
        html = """<div><p>expression: <var>x</var>+<var>y</var>=<var>z</var></p></div>"""
        s = Selector(html)
        assert s.xpath("//var")[0].text == "x"
        assert s.xpath("//var")[0].string == "<var>x</var>"
        assert s.xpath("//var").text == ["x", "y", "z"]
        assert s.xpath("//var").string == ["<var>x</var>", "<var>y</var>", "<var>z</var>"]

    def test_node_selection(self):
        html = """<p></p><p class='primary'>primary</p><p class="minor gray">minor</p>"""
        s = Selector(html)
        primary = s.xpath("//p[@class='primary']")
        assert len(primary) == 1 and primary[0].text == 'primary'
        minor = s.xpath("//p[@class='minor']")
        assert len(minor) == 0
        minor = s.xpath("//p[@class='gray minor']")
        assert len(minor) == 0
        minor = s.xpath("//p[@class='minor gray']")
        assert len(minor) == 1 and minor[0].text == 'minor'
        minor = s.xpath("//p[contains(@class, 'minor')]")
        assert len(minor) == 1 and minor[0].text == 'minor'

    def test_encoding_detection(self):
        html = "<html lang=en><head>" \
               "<title>测试</title>" \
               "<meta charset=gbk>" \
               "<meta http-equiv=Content-Type content='text/html; charset=gbk' />" \
               "</head></html>"
        body = html.encode("gbk")
        with pytest.raises(UnicodeDecodeError):
            Selector(body.decode("utf-8"))
        s = Selector(body.decode("ascii", errors="ignore"))
        assert s.xpath("//meta/@charset")[0].text == "gbk"
        content_type = s.xpath("//meta[@http-equiv='Content-Type']/@content")[0].text
        assert content_type == "text/html; charset=gbk"
        mimetype = parse_mimetype(content_type)
        assert mimetype.type == "text" and mimetype.subtype == "html" and mimetype.parameters.get("charset") == "gbk"

    def test_wrong_arguments(self):
        html = b"<html></html>"
        with pytest.raises(TypeError):
            Selector(html)
        with pytest.raises(ValueError):
            Selector()

    def test_node_context(self):
        html = "<p>header</p><div><p>text</p></div>"
        s = Selector(html)
        assert s.xpath("/p") == []
        assert s.xpath("//p") != []
        assert s.xpath("/html/body/p") != []
        assert s.xpath("//p")[0].text == "header"
        assert s.xpath("//div").xpath("//p")[0].text == "header"
        assert s.xpath("//div").xpath(".//p")[0].text == "text"
        assert s.xpath("//div").xpath("./p")[0].text == "text"


class TestCssSelector:
    def test_selector_list(self):
        html = """<li>a</li><div><ul><li>b</li><li>c</li></ul></div><ul><li>d</li></ul>"""
        s = Selector(html)
        assert s.css("li")[0].text == "a"
        assert s.css("li")[-1].text == "d"
        assert s.css("div li").text[0] == "b"
        assert s.css("div li").text[-1] == "c"

    def test_node_selection(self):
        s = Selector("""<html>
                <body>
                    <h1 class='header'><a href="http://example.com/">example</a></h1>
                    <ul>
                        <li><a class='primary link' href="http://example.com/index">index</a></li>
                        <li><a class='link' href="http://example.com/content">content</a></li>
                    </ul>
                </body>
                </html>""")
        assert s.css('a')[0].text == 'example'
        assert s.css('a[class=link]')[0].text == 'content'
        assert s.css('a[class~=link]')[0].text == 'index'
        assert s.css('ul>a') == []
        assert len(s.css('li>a')) == 2
        assert s.css('ul a')[0].text == 'index'
        assert s.css('ul').css('a')[0].text == 'index'
        assert s.css('li>a:not([class~=primary])')[0].text == 'content'

    def test_attribute_selection(self):
        s = Selector("""<html>
                        <body>
                            <h1 class='header'><a href="http://example.com/">example</a></h1>
                            <ul>
                                <li><a class='primary link' href="http://example.com/">index</a></li>
                                <li><a class='link' href="http://example.com/content">content</a></li>
                            </ul>
                        </body>
                        </html>""")
        assert s.css('h1').xpath('@class')[0].text == 'header'
        assert s.css('li>a[class=link]').xpath('@href')[0].text == 'http://example.com/content'
        assert s.css('h1').attr('class')[0] == 'header'
        assert s.css('h1')[0].attr('class') == 'header'
        assert s.css('li>a').attr('class')[0] == 'primary link'
        assert s.css('li>a').attr('class')[-1] == 'link'
        assert s.css('ul').attr('class') == [None]
        assert s.css('ul')[0].attr('class') is None
        assert s.css('h2').attr('class') == []


class TestXmlText:
    def test_text_selection(self):
        xml = "<xml><p>header</p><div><p>text</p></div></xml>"
        xs = Selector(xml, text_type='xml')
        assert xs.xpath('/p') == []
        assert xs.xpath('//p') != []
        assert xs.xpath('/xml/p') != []
        assert xs.xpath("//p")[0].text == "header"
        assert xs.xpath("//div").xpath("//p")[0].text == "header"
        assert xs.xpath("//div").xpath(".//p")[0].text == "text"
        assert xs.xpath("//div").xpath("./p")[0].text == "text"

    def test_not_xml(self):
        from lxml.etree import XMLSyntaxError
        xml = """expression: <var>x</var>+<var>y</var>=<var>z</var>"""
        with pytest.raises(XMLSyntaxError):
            xs = Selector(xml, text_type='xml')


def test_selection_error():
    xml = """
    <bookstore>
      <book>
        <title lang="eng">Harry Potter</title>
        <price>29.99</price>
      </book>
      <book>
        <title lang="eng">Learning XML</title>
        <price>39.95</price>
      </book>
    </bookstore>"""

    s = Selector(xml, text_type='xml')
    assert s.xpath('/bookstore/book[first()]') == []
    assert s.css('book >> price') == []



# coding=utf-8

project_name = 'quotes'
project_description = ''

# Configure spider
spider = 'quotes.spider.QuotesSpider'

# Configure item pipelines
item_pipelines = [
    'quotes.pipelines.QuotesPipeline'
]



# coding=utf-8

import socket
import struct

_socket_init = socket.socket.__init__


def socket_init(self, *args, **kw):
    _socket_init(self, *args, **kw)
    self.setsockopt(socket.SOL_SOCKET, socket.SO_LINGER, struct.pack("ii", 1, 0))


socket.socket.__init__ = socket_init



# coding=utf-8

from xpaw.http import HttpRequest, HttpResponse


def test_copy_http_request():
    req = HttpRequest('http://example.com/', params={'key': 'value'}, meta={'depth': 0})
    copy_req = req.copy()
    assert copy_req.url == req.url
    assert 'key' in copy_req.params and copy_req.params['key'] == 'value'
    assert 'depth' in copy_req.meta and copy_req.meta['depth'] == 0
    req.meta['depth'] = 1
    assert req.meta['depth'] == 1 and copy_req.meta['depth'] == 0


def test_replace_http_request():
    req = HttpRequest('http://example.com/', 'POST', body=b'body1')
    new_req = req.replace(url='https://example.com/', body=b'body2')
    assert new_req.url == 'https://example.com/'
    assert new_req.body == b'body2'
    assert new_req.method == 'POST'


def test_copy_http_response():
    resp = HttpResponse('http://example.com/', 200, body=b'body')
    copy_resp = resp.copy()
    assert copy_resp.url == 'http://example.com/'
    assert copy_resp.status == 200
    assert copy_resp.body == b'body'


def test_replace_http_response():
    resp = HttpResponse('http://example.com/', 200, body=b'body1')
    new_resp = resp.replace(url='https://example.com/', body=b'body2')
    assert new_resp.url == 'https://example.com/'
    assert new_resp.status == 200
    assert new_resp.body == b'body2'



# coding=utf-8


cluster_start = object()
cluster_shutdown = object()

request_scheduled = object()
request_ignored = object()
response_received = object()

item_scraped = object()
item_ignored = object()



# coding=utf-8

import pytest

from xpaw.spider import SpiderMiddlewareManager, Spider, every
from xpaw.eventbus import EventBus
from xpaw.config import Config
from xpaw import events


class FooSpidermw:
    def __init__(self, d):
        self.d = d

    def open(self):
        self.d['open'] = ''

    def close(self):
        self.d['close'] = ''

    def handle_start_requests(self, result):
        res = []
        for r in result:
            self.d['handle_start_requests'] = r
            res.append(r)
        return res

    def handle_input(self, response):
        self.d['handle_input'] = response

    def handle_output(self, response, result):
        res = []
        for r in result:
            self.d['handle_output'] = (response, r)
            res.append(r)
        return res

    def handle_error(self, response, error):
        self.d['handle_error'] = (response, error)


class DummySpidermw:
    """
    no method
    """


class FooAsyncSpiderMw(FooSpidermw):
    @classmethod
    def from_cluster(cls, cluster):
        return cls(cluster.config['data'])

    async def handle_start_requests(self, result):
        res = []
        for r in result:
            self.d['async_handle_start_requests'] = r
            res.append(r)
        return res

    async def handle_input(self, response):
        self.d['async_handle_input'] = response

    async def handle_output(self, response, result):
        res = []
        for r in result:
            self.d['async_handle_output'] = (response, r)
            res.append(r)
        return res

    async def handle_error(self, response, error):
        self.d['async_handle_error'] = (response, error)


class Cluster:
    def __init__(self, **kwargs):
        self.event_bus = EventBus()
        self.config = Config(kwargs)


async def test_spider_middleware_manager_handlers():
    data = {}
    cluster = Cluster(spider_middlewares=[lambda d=data: FooSpidermw(d),
                                          DummySpidermw,
                                          FooAsyncSpiderMw],
                      spider_middlewares_base=None,
                      data=data)
    spidermw = SpiderMiddlewareManager.from_cluster(cluster)
    response_obj = object()
    result_obj = object()
    error_obj = object()
    await cluster.event_bus.send(events.cluster_start)
    await spidermw._handle_start_requests((result_obj,))
    await spidermw._handle_input(response_obj)
    await spidermw._handle_output(response_obj, (result_obj,))
    await spidermw._handle_error(response_obj, error_obj)
    await cluster.event_bus.send(events.cluster_shutdown)
    assert 'open' in data and 'close' in data
    assert data['handle_start_requests'] is result_obj
    assert data['handle_input'] is response_obj
    assert data['handle_output'][0] is response_obj and data['handle_output'][1] is result_obj
    assert data['handle_error'][0] is response_obj and data['handle_error'][1] is error_obj
    assert data['async_handle_start_requests'] is result_obj
    assert data['async_handle_input'] is response_obj
    assert data['async_handle_output'][0] is response_obj and data['async_handle_output'][1] is result_obj
    assert data['async_handle_error'][0] is response_obj and data['async_handle_error'][1] is error_obj

    data2 = {}
    cluster2 = Cluster(spider_middlewares={lambda d=data2: FooSpidermw(d): 0},
                       spider_middlewares_base=None,
                       data=data2)
    spidermw2 = SpiderMiddlewareManager.from_cluster(cluster2)
    response_obj2 = object()
    await cluster2.event_bus.send(events.cluster_start)
    await spidermw2._handle_input(response_obj2)
    await cluster2.event_bus.send(events.cluster_shutdown)
    assert 'open' in data2 and 'close' in data2
    assert data2['handle_input'] is response_obj2

    cluster3 = Cluster(spider_middlewares=None, spider_middlewares_base=None, data={})
    SpiderMiddlewareManager.from_cluster(cluster3)


class FooSpider(Spider):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.data = self.config['data']

    def open(self):
        super().open()
        self.data['open'] = ''

    def close(self):
        super().close()
        self.data['close'] = ''


async def test_spider(loop):
    data = {}
    cluster = Cluster(data=data, loop=loop)
    spider = FooSpider.from_cluster(cluster)
    await cluster.event_bus.send(events.cluster_start)
    with pytest.raises(NotImplementedError):
        spider.start_requests()
    with pytest.raises(NotImplementedError):
        spider.parse(None)
    await cluster.event_bus.send(events.cluster_shutdown)
    assert 'open' in data and 'close' in data


def test_every_value_error():
    with pytest.raises(ValueError):
        @every()
        def func():
            pass


def test_every():
    @every(hours=1)
    def func_hours():
        pass

    assert func_hours.cron_job is True and func_hours.cron_tick == 3600

    @every(minutes=1)
    def func_minutes():
        pass

    assert func_minutes.cron_job is True and func_minutes.cron_tick == 60

    @every(seconds=1)
    def func_seconds():
        pass

    assert func_seconds.cron_job is True and func_seconds.cron_tick == 1

    @every(hours=1, minutes=1, seconds=1)
    def func():
        pass

    assert func.cron_job is True and func.cron_tick == 3661



# coding=utf-8

import logging
import inspect
from asyncio import CancelledError

from .middleware import MiddlewareManager
from . import events
from .utils import iterable_to_list
from .http import HttpRequest
from .errors import IgnoreRequest

log = logging.getLogger(__name__)


class Spider:
    def __init__(self, config=None, **kwargs):
        self.config = config
        self.__dict__.update(kwargs)

    @classmethod
    def from_cluster(cls, cluster):
        spider = cls(config=cluster.config, cluster=cluster)
        cluster.event_bus.subscribe(spider.open, events.cluster_start)
        cluster.event_bus.subscribe(spider.close, events.cluster_shutdown)
        return spider

    @property
    def logger(self):
        return log

    def log(self, message, *args, level=logging.INFO, **kwargs):
        self.logger.log(level, message, *args, **kwargs)

    def parse(self, response):
        raise NotImplementedError

    def start_requests(self):
        raise NotImplementedError

    def open(self):
        pass

    def close(self):
        pass

    async def request_success(self, response):
        callback = response.request.callback
        if callback:
            res = self._parse_method(callback)(response)
        else:
            res = self.parse(response)
        if inspect.iscoroutine(res):
            res = await res
        return res

    async def request_error(self, request, error):
        log.warning('%s is failed: %s', request, error, exc_info=None if isinstance(error, IgnoreRequest) else True)
        try:
            if request and request.errback:
                r = self._parse_method(request.errback)(request, error)
                if inspect.iscoroutine(r):
                    await r
        except CancelledError:
            raise
        except Exception:
            log.warning("Error occurred in error callback", exc_info=True)

    def _parse_method(self, method):
        if isinstance(method, str):
            method = getattr(self, method)
        return method


def _isiterable(obj):
    return hasattr(obj, "__iter__") or hasattr(obj, "__aiter__")


class SpiderMiddlewareManager(MiddlewareManager):
    def __init__(self, *middlewares):
        self._input_handlers = []
        self._output_handlers = []
        self._error_handlers = []
        self._start_requests_handlers = []
        super().__init__(*middlewares)

    def _add_middleware(self, middleware):
        super()._add_middleware(middleware)
        if hasattr(middleware, "handle_input"):
            self._input_handlers.append(middleware.handle_input)
        if hasattr(middleware, "handle_output"):
            self._output_handlers.insert(0, middleware.handle_output)
        if hasattr(middleware, "handle_error"):
            self._error_handlers.insert(0, middleware.handle_error)
        if hasattr(middleware, "handle_start_requests"):
            self._start_requests_handlers.insert(0, middleware.handle_start_requests)

    @classmethod
    def _middleware_list_from_config(cls, config):
        return cls._make_component_list('spider_middlewares', config)

    async def parse(self, spider, response):
        request = response.request
        try:
            try:
                await self._handle_input(response)
            except CancelledError:
                raise
            except Exception as e:
                await spider.request_error(request, e)
                raise e
            res = await spider.request_success(response)
            assert res is None or _isiterable(res), \
                "Parsing result must be None or an iterable object, got {}".format(type(res).__name__)
            result = await iterable_to_list(res)
        except CancelledError:
            raise
        except Exception as e:
            res = await self._handle_error(response, e)
            if isinstance(res, Exception):
                raise res
            result = await iterable_to_list(res)
        if result:
            res = await self._handle_output(response, result)
            result = await iterable_to_list(res)
        return result

    async def start_requests(self, spider):
        res = spider.start_requests()
        if inspect.iscoroutine(res):
            res = await res
        assert res is None or _isiterable(res), \
            "Start requests must be None or an iterable object, got {}".format(type(res).__name__)
        result = await iterable_to_list(res)
        if result:
            res = await self._handle_start_requests(result)
            result = await iterable_to_list(res)
        return result

    async def _handle_input(self, response):
        for method in self._input_handlers:
            res = method(response)
            if inspect.iscoroutine(res):
                res = await res
            assert res is None, \
                "Input handler must return None, got {}".format(type(res).__name__)

    async def _handle_output(self, response, result):
        for method in self._output_handlers:
            result = method(response, result)
            if inspect.iscoroutine(result):
                result = await result
            assert _isiterable(result), \
                "Output handler must return an iterable object, got {}".format(type(result).__name__)
        return result

    async def _handle_error(self, response, error):
        for method in self._error_handlers:
            res = method(response, error)
            if inspect.iscoroutine(res):
                res = await res
            assert res is None or _isiterable(res), \
                "Exception handler must return None or an iterable object, got {}".format(type(res).__name__)
            if res is not None:
                return res
        return error

    async def _handle_start_requests(self, result):
        for method in self._start_requests_handlers:
            result = method(result)
            if inspect.iscoroutine(result):
                result = await result
            assert _isiterable(result), \
                "Start requests handler must return an iterable object, got {}".format(type(result).__name__)
        return result


def every(hours=None, minutes=None, seconds=None):
    def wrapper(func):
        func.cron_job = True
        func.cron_tick = hours * 3600 + minutes * 60 + seconds
        return func

    if hours is None and minutes is None and seconds is None:
        raise ValueError('At least one of the parameters (hours, minutes and seconds) is not none')
    if hours is None:
        hours = 0
    if minutes is None:
        minutes = 0
    if seconds is None:
        seconds = 0
    return wrapper


class RequestsSpider(Spider):
    def start_requests(self):
        requests = self.config.get('start_requests')
        i = 0
        for r in requests:
            if isinstance(r, str):
                r = HttpRequest(r)
            if isinstance(r, HttpRequest):
                r.meta['request_index'] = i
                r.dont_filter = True
                r.callback = self.parse
                r.errback = self.handle_error
                yield r
            else:
                self.logger.warning('Requests must be str or HttpRequest, got %s', type(r).__name__)
            i += 1

    def parse(self, response):
        results = self.config.get('results')
        results[response.meta['request_index']] = response

    def handle_error(self, request, err):
        results = self.config.get('results')
        results[request.meta['request_index']] = err
